# Alex Martelli's 'Borg' singleton implementation:
# http://python-3-patterns-idioms-test.readthedocs.io/en/latest/Singleton.html#id1
#
# Modified (and a bit improved/enhanced) by Adam Karpierz, 2014

from typing import Any
from typing_extensions import Self

__all__ = ('Borg',)


class Borg:

    __shared_state: dict[str, Any] = {}

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        """Constructor"""
        self = super().__new__(cls)
        self.__dict__ = Borg.__shared_state
        return self



# Copyright (c) 1994 Adam Karpierz
# SPDX-License-Identifier: Zlib

__all__ = ('arch',)


def get_python_arch() -> str | None:
    import sys
    import platform
    machine = platform.machine().lower()
    is_32bits = (sys.maxsize <= 2**32)

    # x86
    if machine in ("x86_64", "amd64"):
        arch = "x86_64" if not is_32bits else "x86"
    elif machine in ("i386", "i686", "x86", "ia32"):
        arch = "x86"
    # ARM
    elif machine in ("arm64",):  # macOS never reports aarch64
        arch = "arm64" if not is_32bits else "arm"
    else:
        arch = None  # unknown or unsupported

    return arch


arch = get_python_arch()


def macos_version() -> tuple[int, ...]:
    import platform
    version  = [int(x) for x in platform.mac_ver()[0].split(".")]
    version += [0] * (3 - len(version))
    return tuple(version)



# Copyright (c) 2018 Adam Karpierz
# SPDX-License-Identifier: Zlib

import sys
import types
import inspect
from pathlib import Path

__all__ = ('module_path',)


def module_path(module: types.ModuleType | None = None,
                *, level: int = 1) -> Path:
    if module is not None:
        mfile = inspect.getfile(module)
    else:
        frame = sys._getframe(level)
        module = inspect.getmodule(frame)
        if module is not None:
            mfile = inspect.getfile(module)
        else:
            mfile = frame.f_globals["__file__"]
    return Path(mfile).resolve().parent



# flake8-in-file-ignores: noqa: E305,F401

# Copyright (c) 1994 Adam Karpierz
# SPDX-License-Identifier: Zlib

import ctypes
from ctypes import windll
from ctypes import wintypes
from ctypes import WINFUNCTYPE
from ctypes.wintypes import (
    CHAR, WCHAR, BOOLEAN, BOOL, BYTE, WORD, DWORD, SHORT, USHORT, INT,
    UINT, LONG, ULONG, LARGE_INTEGER, ULARGE_INTEGER, FLOAT, DOUBLE,
    LPBYTE, PBYTE, LPWORD, PWORD, LPDWORD, PDWORD, LPLONG, PLONG, LPSTR,
    LPWSTR, LPCSTR, LPVOID, LPCVOID, LPVOID as PVOID, HANDLE, LPHANDLE,
    PHANDLE, HMODULE, HLOCAL, WPARAM, LPARAM, FILETIME, LPFILETIME,
)

from ctypes.wintypes import WPARAM as ULONG_PTR  # workaround
PULONG_PTR = ctypes.POINTER(ULONG_PTR)

ULONG32   = ctypes.c_uint32
ULONGLONG = ctypes.c_uint64
DWORDLONG = ctypes.c_uint64
SIZE_T    = ctypes.c_size_t

WAIT_ABANDONED = 0x00000080
WAIT_OBJECT_0  = 0x00000000
WAIT_TIMEOUT   = 0x00000102
WAIT_FAILED    = 0xFFFFFFFF

IGNORE   = 0
INFINITE = 0xFFFFFFFF

FORMAT_MESSAGE_ALLOCATE_BUFFER = 0x00000100
FORMAT_MESSAGE_ARGUMENT_ARRAY  = 0x00002000
FORMAT_MESSAGE_FROM_HMODULE    = 0x00000800
FORMAT_MESSAGE_FROM_STRING     = 0x00000400
FORMAT_MESSAGE_FROM_SYSTEM     = 0x00001000
FORMAT_MESSAGE_IGNORE_INSERTS  = 0x00000200
FORMAT_MESSAGE_MAX_WIDTH_MASK  = 0x000000FF

GetActiveProcessorCount = windll.kernel32.GetActiveProcessorCount
GetActiveProcessorCount.restype  = DWORD
GetActiveProcessorCount.argtypes = [WORD]

GetModuleFileNameW = windll.kernel32.GetModuleFileNameW
GetModuleFileNameW.restype  = DWORD
GetModuleFileNameW.argtypes = [HMODULE,
                               LPWSTR,
                               DWORD]

GetProcessId = windll.kernel32.GetProcessId
GetProcessId.restype  = DWORD
GetProcessId.argtypes = [HANDLE]

GetCurrentProcess = windll.kernel32.GetCurrentProcess
GetCurrentProcess.restype  = HANDLE
GetCurrentProcess.argtypes = []

GetCurrentProcessId = windll.kernel32.GetCurrentProcessId
GetCurrentProcessId.restype  = DWORD
GetCurrentProcessId.argtypes = []

GetProcessTimes = windll.kernel32.GetProcessTimes
GetProcessTimes.restype  = BOOL
GetProcessTimes.argtypes = [HANDLE,
                            LPFILETIME,
                            LPFILETIME,
                            LPFILETIME,
                            LPFILETIME]

OpenProcessToken = windll.advapi32.OpenProcessToken
OpenProcessToken.restype  = BOOL
OpenProcessToken.argtypes = [HANDLE,   # ProcessHandle
                             DWORD,    # DesiredAccess
                             PHANDLE]  # TokenHandle

TOKEN_INFORMATION_CLASS  = DWORD
PTOKEN_INFORMATION_CLASS = PDWORD

GetTokenInformation = windll.advapi32.GetTokenInformation
GetTokenInformation.restype  = BOOL
GetTokenInformation.argtypes = [HANDLE,  # TokenHandle
                                TOKEN_INFORMATION_CLASS,  # TokenInformationClass
                                LPVOID,  # TokenInformation
                                DWORD,   # TokenInformationLength
                                PDWORD]  # ReturnLength

GetExitCodeProcess = windll.kernel32.GetExitCodeProcess
GetExitCodeProcess.restype  = BOOL
GetExitCodeProcess.argtypes = [HANDLE, LPDWORD]

ExitProcess = windll.kernel32.ExitProcess
ExitProcess.restype  = None
ExitProcess.argtypes = [UINT]

class SECURITY_ATTRIBUTES(ctypes.Structure):
    _fields_ = [
    ("nLength",              DWORD),
    ("lpSecurityDescriptor", LPVOID),
    ("bInheritHandle",       BOOL),
]
LPSECURITY_ATTRIBUTES = ctypes.POINTER(SECURITY_ATTRIBUTES)

LPTHREAD_START_ROUTINE = WINFUNCTYPE(DWORD, LPVOID)
CreateThread = windll.kernel32.CreateThread
CreateThread.restype  = HANDLE
CreateThread.argtypes = [LPSECURITY_ATTRIBUTES,
                         SIZE_T,
                         LPTHREAD_START_ROUTINE,
                         LPVOID,
                         DWORD,
                         LPDWORD]

GetCurrentThreadId = windll.kernel32.GetCurrentThreadId
GetCurrentThreadId.restype  = DWORD
GetCurrentThreadId.argtypes = []

WaitForSingleObject = windll.kernel32.WaitForSingleObject
WaitForSingleObject.restype  = DWORD
WaitForSingleObject.argtypes = [HANDLE,
                                DWORD]

SetEvent = windll.kernel32.SetEvent
SetEvent.restype  = BOOL
SetEvent.argtypes = [HANDLE]

CreateSemaphore = windll.kernel32.CreateSemaphoreA
CreateSemaphore.restype  = HANDLE
CreateSemaphore.argtypes = [LPSECURITY_ATTRIBUTES,
                            LONG,
                            LONG,
                            LPCSTR]

ReleaseSemaphore = windll.kernel32.ReleaseSemaphore
ReleaseSemaphore.restype  = BOOL
ReleaseSemaphore.argtypes = [HANDLE,
                             LONG,
                             LPLONG]

Sleep = windll.kernel32.Sleep
Sleep.restype  = None
Sleep.argtypes = [DWORD]

GetStdHandle = windll.kernel32.GetStdHandle
GetStdHandle.restype  = HANDLE
GetStdHandle.argtypes = [DWORD]

DuplicateHandle = windll.kernel32.DuplicateHandle
DuplicateHandle.restype  = BOOL
DuplicateHandle.argtypes = [HANDLE,
                            HANDLE,
                            HANDLE,
                            LPHANDLE,
                            DWORD,
                            BOOL,
                            DWORD]

SetHandleInformation = windll.kernel32.SetHandleInformation
SetHandleInformation.restype  = BOOL
SetHandleInformation.argtypes = [HANDLE,
                                 DWORD,
                                 DWORD]

CloseHandle = windll.kernel32.CloseHandle
CloseHandle.restype  = BOOL
CloseHandle.argtypes = [HANDLE]

# DWORD WINAPI GetLastError(void);
GetLastError = windll.kernel32.GetLastError
GetLastError.restype  = DWORD
GetLastError.argtypes = []

FormatMessageA = windll.kernel32.FormatMessageA
FormatMessageA.restype  = DWORD
FormatMessageA.argtypes = [DWORD,
                           LPCVOID,
                           DWORD,
                           DWORD,
                           LPSTR,
                           DWORD,
                           LPVOID]

class SYSTEMTIME(ctypes.Structure):
    _fields_ = [
    ("wYear",         WORD),
    ("wMonth",        WORD),
    ("wDayOfWeek",    WORD),
    ("wDay",          WORD),
    ("wHour",         WORD),
    ("wMinute",       WORD),
    ("wSecond",       WORD),
    ("wMilliseconds", WORD),
]
LPSYSTEMTIME = ctypes.POINTER(SYSTEMTIME)

GetLocalTime = windll.kernel32.GetLocalTime
GetLocalTime.restype  = None
GetLocalTime.argtypes = [LPSYSTEMTIME]

SetLocalTime = windll.kernel32.SetLocalTime
SetLocalTime.restype  = BOOL
SetLocalTime.argtypes = [LPSYSTEMTIME]

GetSystemTime = windll.kernel32.GetSystemTime
GetSystemTime.restype  = None
GetSystemTime.argtypes = [LPSYSTEMTIME]

SetSystemTime = windll.kernel32.SetSystemTime
SetSystemTime.restype  = BOOL
SetSystemTime.argtypes = [LPSYSTEMTIME]

# http://posted-stuff.blogspot.com/2009/07/iocp-based-sockets-with-ctypes-in_22.html

WSADESCRIPTION_LEN = 256
WSASYS_STATUS_LEN  = 128

class WSADATA(ctypes.Structure):
    _fields_ = [
    ("wVersion",       WORD),
    ("wHighVersion",   WORD),
    ("szDescription",  ctypes.c_char * (WSADESCRIPTION_LEN + 1)),
    ("szSystemStatus", ctypes.c_char * (WSASYS_STATUS_LEN  + 1)),
    ("iMaxSockets",    ctypes.c_ushort),
    ("iMaxUdpDg",      ctypes.c_ushort),
    ("lpVendorInfo",   ctypes.c_char_p),
]

LPWSADATA = ctypes.POINTER(WSADATA)

WSAStartup = windll.Ws2_32.WSAStartup
WSAStartup.restype  = ctypes.c_int
WSAStartup.argtypes = [WORD, LPWSADATA]

WSACleanup = windll.Ws2_32.WSACleanup
WSACleanup.restype  = ctypes.c_int
WSACleanup.argtypes = []

CP_ACP        = 0      # default to ANSI code page
CP_OEMCP      = 1      # default to OEM  code page
CP_MACCP      = 2      # default to MAC  code page
CP_THREAD_ACP = 3      # current thread's ANSI code page
CP_SYMBOL     = 42     # SYMBOL translations

CP_UTF7       = 65000  # UTF-7 translation
CP_UTF8       = 65001  # UTF-8 translation

# UINT WINAPI GetConsoleCP(void);
GetConsoleCP = windll.kernel32.GetConsoleCP
GetConsoleCP.restype  = UINT
GetConsoleCP.argtypes = []

# UINT WINAPI GetConsoleOutputCP(void);
GetConsoleOutputCP = windll.kernel32.GetConsoleOutputCP
GetConsoleOutputCP.restype  = UINT
GetConsoleOutputCP.argtypes = []

# BOOL WINAPI SetConsoleCP(
#  _In_  UINT wCodePageID
# );
SetConsoleCP = windll.kernel32.SetConsoleCP
SetConsoleCP.restype  = BOOL
SetConsoleCP.argtypes = [UINT]

# BOOL WINAPI SetConsoleOutputCP(
#  _In_  UINT wCodePageID
# );
SetConsoleOutputCP = windll.kernel32.SetConsoleOutputCP
SetConsoleOutputCP.restype  = BOOL
SetConsoleOutputCP.argtypes = [UINT]

PHANDLER_ROUTINE = WINFUNCTYPE(BOOL, DWORD)
SetConsoleCtrlHandler = windll.kernel32.SetConsoleCtrlHandler
SetConsoleCtrlHandler.restype  = BOOL
SetConsoleCtrlHandler.argtypes = [PHANDLER_ROUTINE,
                                  BOOL]

LocalFree = windll.kernel32.LocalFree
LocalFree.restype  = HLOCAL
LocalFree.argtypes = [HLOCAL]

MAKEWORD = lambda blow, bhigh: (bhigh << 8) + blow

del ctypes



# Copyright (c) 2012 Adam Karpierz
# SPDX-License-Identifier: Zlib

from typing import TypeAlias, Any
from collections.abc import Callable

__all__ = ('deprecated',)

AnyCallable: TypeAlias = Callable[..., Any]


def deprecated(func: AnyCallable) -> AnyCallable:
    """This is a decorator which can be used to mark functions as deprecated.

    It will result in a warning being emitted when the function is used.
    """
    from functools import wraps
    from warnings  import warn

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        warn("Call to deprecated function '{0}' ({1}:{2}).".format(
             func.__name__,
             func.__code__.co_filename,
             func.__code__.co_firstlineno + 1),
             category=DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)

    return wrapper



# Copyright (c) 2016 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
from unittest import mock
import tempfile
import shutil
import sys
import os
import time
import stat
import re
from pathlib import Path as StdPath

import utlx
from utlx.epath import Path


class TestPath(unittest.TestCase):

    def setUp(self):
        self.temp_dir = Path(tempfile.mkdtemp())
        self.file = self.temp_dir / "test.txt"
        self.file.write_text("Hello World\nHello Python\n")

    def tearDown(self):
        shutil.rmtree(str(self.temp_dir), ignore_errors=True)

    def test_exists_and_mkdir(self):
        new_dir = self.temp_dir / "new"
        self.assertFalse(new_dir.exists())
        new_dir.mkdir()
        self.assertTrue(new_dir.exists())

    def test_rmdir_and_cleardir(self):
        subdir = self.temp_dir / "sub"
        subdir.mkdir()
        (subdir / "file.txt").write_text("data")
        (subdir / "file_readonly.txt").write_text("data_readonly")
        (subdir / "file_readonly.txt").chmod(stat.S_IREAD)
        (subdir / "sub_sub").mkdir()
        (subdir / "sub_sub" / "file.txt").write_text("data_sub")
        (subdir / "sub_sub" / "file_readonly.txt").write_text("data_sub_readonly")
        (subdir / "sub_sub" / "file_readonly.txt").chmod(stat.S_IREAD)
        subdir.cleardir()
        self.assertTrue(subdir.exists())
        self.assertEqual(list(subdir.iterdir()), [])

        (subdir / "file.txt").write_text("data")
        (subdir / "file_readonly.txt").write_text("data_readonly")
        (subdir / "file_readonly.txt").chmod(stat.S_IREAD)
        (subdir / "sub_sub").mkdir()
        (subdir / "sub_sub" / "file.txt").write_text("data_sub")
        (subdir / "sub_sub" / "file_readonly.txt").write_text("data_sub_readonly")
        (subdir / "sub_sub" / "file_readonly.txt").chmod(stat.S_IREAD)
        subdir.rmdir()
        self.assertFalse(subdir.exists())

        subdir = self.temp_dir / "nonexistent.txt"
        self.assertFalse(subdir.exists())
        subdir.cleardir()

        subdir = self.file
        self.assertTrue(subdir.exists())
        with self.assertRaises(NotADirectoryError):
            subdir.cleardir()

    def test_rmdir_on_nonexistent_path_does_nothing(self):
        ghost = self.temp_dir / "ghost"
        try:
            ghost.rmdir()
        except Exception as e:  # pragma: no cover
            self.fail(f"rmdir raised {e} unexpectedly")

    def test_cleardir_on_symlink_raises(self):
        target = self.temp_dir / "target"
        target.mkdir()
        symlink = self.temp_dir / "link"
        symlink.symlink_to(target, target_is_directory=True)
        with self.assertRaises(NotADirectoryError):
            symlink.cleardir()

    def test_cleardir_on_symlink_when_has_file_attrs_is_false_raises(self):
        """Ensure Path.exists() uses the else branch of _is_real_link \
           when _HAS_FILE_ATTRS is False."""
        target = self.temp_dir / "target_else"
        target.mkdir()
        symlink = self.temp_dir / "link_else"
        symlink.symlink_to(target, target_is_directory=True)
        with mock.patch("utlx.epath._HAS_FILE_ATTRS", False):
            with self.assertRaises(NotADirectoryError):
                symlink.cleardir()

    def test_copydir_basic(self):
        src = self.temp_dir / "src_basic"
        src.mkdir()
        (src / "file1.txt").write_text("Hello")
        (src / "file2.txt").write_text("World")
        dst = self.temp_dir / "dst_basic"

        copied = src.copydir(dst)
        self.assertTrue((copied / "file1.txt").exists())
        self.assertTrue((copied / "file2.txt").exists())
        self.assertEqual((copied / "file1.txt").read_text(), "Hello")
        self.assertEqual((copied / "file2.txt").read_text(), "World")

    def test_copydir_empty_directory(self):
        src = self.temp_dir / "src_empty"
        src.mkdir()
        dst = self.temp_dir / "dst_empty"

        copied = src.copydir(dst)
        self.assertTrue(copied.exists())
        self.assertTrue(copied.is_dir())
        self.assertEqual(list(copied.iterdir()), [])

    def test_copydir_nested_empty_directories(self):
        src = self.temp_dir / "src_nested"
        (src / "a" / "b" / "c").mkdir(parents=True)
        dst = self.temp_dir / "dst_nested"

        copied = src.copydir(dst)
        self.assertTrue((copied / "a" / "b" / "c").exists())
        self.assertTrue((copied / "a" / "b" / "c").is_dir())
        self.assertEqual(list((copied / "a" / "b" / "c").iterdir()), [])

    def test_copydir_with_ignore(self):
        src = self.temp_dir / "src_ignore"
        src.mkdir()
        (src / "file1.txt").write_text("data")
        (src / "file2.txt").write_text("data")
        dst = self.temp_dir / "dst_ignore"

        def ignore_func(dir, files):
            return ["file2.txt"]

        copied = src.copydir(dst, ignore=ignore_func)
        self.assertTrue(copied.exists())
        self.assertTrue((copied / "file1.txt").exists())
        self.assertFalse((copied / "file2.txt").exists())

    def test_copydir_ignore_all(self):
        src = self.temp_dir / "src_ignore_all"
        src.mkdir()
        (src / "file1.txt").write_text("data")
        (src / "file2.txt").write_text("data")
        dst = self.temp_dir / "dst_ignore_all"

        def ignore_all(dir, files):
            return files  # ignore all

        copied = src.copydir(dst, ignore=ignore_all)
        self.assertTrue(copied.exists())
        self.assertEqual(list(copied.iterdir()), [])

    def test_copydir_with_symlinks(self):
        src = self.temp_dir / "src_symlink"
        src.mkdir()
        target_file = src / "target.txt"
        target_file.write_text("symlinked")
        symlink = src / "link.txt"
        symlink.symlink_to(target_file)
        dst = self.temp_dir / "dst_symlink"

        copied = src.copydir(dst, symlinks=True)
        copied_link = copied / "link.txt"
        self.assertTrue(copied_link.is_symlink())
        self.assertEqual(copied_link.read_text(), "symlinked")

    def test_copydir_with_custom_copy_function(self):
        src = self.temp_dir / "src_custom"
        src.mkdir()
        (src / "file.txt").write_text("original")
        dst = self.temp_dir / "dst_custom"

        def custom_copy(src_path, dst_path):
            shutil.copy2(src_path, dst_path)
            with open(dst_path, "a") as f:
                f.write(" [copied]")

        copied = src.copydir(dst, copy_function=custom_copy)
        content = (copied / "file.txt").read_text()
        self.assertIn("original", content)
        self.assertIn("[copied]", content)

    def test_copy_and_move(self):
        dst = self.temp_dir / "copy.txt"
        copied = self.file.copy(dst)
        self.assertTrue(dst.exists())
        self.assertEqual(dst.read_text(), "Hello World\nHello Python\n")

        moved_path = self.temp_dir / "moved.txt"
        moved = copied.move(moved_path)
        self.assertTrue(moved_path.exists())
        self.assertFalse(dst.exists())
        self.assertEqual(moved.read_text(), "Hello World\nHello Python\n")

    def test_move_nonexistent(self):
        src = self.temp_dir / "nonexistent.txt"
        self.assertFalse(src.exists())

        moved_path = self.temp_dir / "nonexistent_moved.txt"
        self.assertFalse(moved_path.exists())
        moved = src.move(moved_path)
        self.assertIsNone(moved)
        self.assertFalse(moved_path.exists())

    def test_unlink_and_permission_handling(self):
        self.file.chmod(stat.S_IREAD)
        self.file.unlink()
        self.assertFalse(self.file.exists())

    def test_unlink_missing_ok_true(self):
        ghost = self.temp_dir / "ghost.txt"
        self.assertFalse(ghost.exists())
        ghost.unlink()  # missing_ok default is True
        ghost.unlink(missing_ok=True)

    def test_unlink_missing_ok_false_raises(self):
        ghost = self.temp_dir / "ghost.txt"
        self.assertFalse(ghost.exists())
        with self.assertRaises(FileNotFoundError):
            ghost.unlink(missing_ok=False)

    def test_file_hash(self):
        hash_val = self.file.file_hash("md5")
        self.assertTrue(hasattr(hash_val, "hexdigest"))
        self.assertIsInstance(hash_val.hexdigest(), str)

    def test_file_hash_invalid_algorithm(self):
        with self.assertRaises(ValueError):
            self.file.file_hash("unknownhash")

    def test_dir_hash(self):
        hash_val = self.temp_dir.dir_hash("sha256")
        self.assertTrue(hasattr(hash_val, "hexdigest"))

    def test_dir_hash_empty_directory(self):
        empty_dir = self.temp_dir / "empty"
        empty_dir.mkdir()
        hash_val = empty_dir.dir_hash("md5")
        self.assertTrue(hasattr(hash_val, "hexdigest"))

    def test_unpack_archive(self):
        archive = self.temp_dir / "archive.zip"
        shutil.make_archive(str(archive.with_suffix("")), "zip", str(self.temp_dir))
        extract_dir = self.temp_dir / "extracted"
        archive.unpack_archive(extract_dir)
        self.assertTrue(extract_dir.exists())
        self.assertTrue((extract_dir / "test.txt").exists())

    def test_unpack_archive_invalid_format(self):
        archive = self.temp_dir / "archive.zip"
        shutil.make_archive(str(archive.with_suffix("")), "zip", str(self.temp_dir))
        with self.assertRaises(shutil.ReadError):
            archive.unpack_archive(format="tar")

    def test_chdir_and_pushd(self):
        original = StdPath.cwd()
        self.temp_dir.chdir()
        self.assertEqual(StdPath.cwd(), self.temp_dir)

        with self.temp_dir.pushd():
            self.assertEqual(StdPath.cwd(), self.temp_dir)
        self.assertEqual(StdPath.cwd(), self.temp_dir)

    def test_which(self):
        python_path = Path.which("python")
        self.assertIsNotNone(python_path)
        self.assertTrue(python_path.exists())

    @mock.patch.object(os, 'link', side_effect=NotImplementedError, create=True)
    def test_hardlink_to_not_supported(self, _):
        if sys.version_info[:2] >= (3, 11) or not hasattr(os, "link"):
            with self.assertRaises(NotImplementedError):
                (self.temp_dir / "hard.txt").hardlink_to(self.file)
        else: pass  # pragma: no cover

    def test_copystat_copies_mtime(self):
        with tempfile.NamedTemporaryFile() as src_file, \
             tempfile.NamedTemporaryFile() as dst_file:

            src = Path(src_file.name)
            dst = Path(dst_file.name)

            # We modify the source's modification time.
            new_mtime = time.time() - 3600  # 1 hour earlier
            os.utime(src, (new_mtime, new_mtime))

            # We make sure the timestamps differ before copying
            src_stat = os.stat(src)
            dst_stat = os.stat(dst)
            self.assertNotEqual(src_stat.st_mtime, dst_stat.st_mtime)

            # We copy the metadata
            src.copystat(dst)

            # We check whether the mtime was successfully copied
            dst_stat_after = os.stat(dst)
            self.assertEqual(src_stat.st_mtime, dst_stat_after.st_mtime)

    def test_copystat_symlink_behavior(self):
        with tempfile.NamedTemporaryFile() as src_file, \
             tempfile.NamedTemporaryFile() as dst_file:

            src = Path(src_file.name)
            dst = Path(dst_file.name)
            symlink = dst.parent / (dst.name + "_link")
            symlink.symlink_to(dst)

            # We modify the source's modification time.
            new_mtime = time.time() - 7200  # 2 hours earlier
            os.utime(src, (new_mtime, new_mtime))

            # We copy the metadata to the symlink (without following it)
            src.copystat(symlink, follow_symlinks=False)

            # We check whether the metadata was assigned to the symlink
            # (if the system allows it)
            try:
                link_stat = os.lstat(symlink)
                self.assertEqual(link_stat.st_mtime, new_mtime)
            except AssertionError:
                # Some systems ignore the mtime of symlinks
                pass

        # Cleanup
        os.remove(symlink)

    def test_pushd_restores_directory_on_exception(self):
        original = StdPath.cwd()
        try:
            with self.temp_dir.pushd():
                raise RuntimeError("Simulated error")
        except RuntimeError:
            self.assertEqual(StdPath.cwd(), original)


class TestSedInPlace(unittest.TestCase):

    def setUp(self):
        self.temp_dir = Path(tempfile.mkdtemp())
        self.file = self.temp_dir / "test.txt"
        self.file.write_text("Hello World\nHello Python\n")

    def tearDown(self):
        shutil.rmtree(str(self.temp_dir), ignore_errors=True)

    def test_sed_inplace(self):
        self.file.sed_inplace("Hello", "Hi")
        content = self.file.read_text()
        self.assertIn("Hi World", content)
        self.assertIn("Hi Python", content)

    def test_sed_inplace_with_string_pattern_and_encoding(self):
        self.file.sed_inplace("Hello", "Hi", encoding="utf-8")
        content = self.file.read_text()
        self.assertIn("Hi World", content)
        self.assertIn("Hi Python", content)

    def test_sed_inplace_with_compiled_pattern(self):
        pattern = re.compile("World")
        self.file.sed_inplace(pattern, "Universe", encoding="utf-8")
        content = self.file.read_text()
        self.assertIn("Hello Universe", content)

    def test_sed_inplace_detected_encoding(self):
        fake_detected = mock.Mock()
        fake_detected.encoding = "utf-8"
        fake_detected.__str__ = lambda self: "Hi World\nHi Python\n"
        with mock.patch("utlx.epath.charset_normalizer.from_bytes") as mock_from_bytes:
            mock_from_bytes.return_value.best.return_value = fake_detected
            self.file.sed_inplace("Hello", "Hi")
        content = self.file.read_text()
        self.assertIn("Hi World", content)

    def test_sed_inplace_no_detected_encoding(self):
        with mock.patch("utlx.epath.charset_normalizer.from_bytes") as mock_from_bytes:
            mock_from_bytes.return_value.best.return_value = None
            self.file.sed_inplace("Hello", "Hi")
        content = self.file.read_text()
        self.assertIn("Hi World", content)

    def test_sed_inplace_multiline_flag(self):
        self.file.sed_inplace("^Hello", "Hi", flags=re.MULTILINE, encoding="utf-8")
        content = self.file.read_text()
        self.assertTrue(content.startswith("Hi"))

    def test_sed_inplace_raises_unicode_error(self):
        """Ensure sed_inplace raises UnicodeError when decoding fails."""
        class BadBytes(bytes):
            def decode(self, *args, **kwargs):
                raise Exception("decode fail")

        bad_file = self.temp_dir / "bad.txt"

        # Patch read_bytes to return our BadBytes instance
        with mock.patch.object(type(bad_file), "read_bytes", return_value=BadBytes(b"xxx")), \
             mock.patch("utlx.epath.charset_normalizer.from_bytes") as mock_from_bytes:
            mock_from_bytes.return_value.best.return_value = None
            with self.assertRaises(UnicodeError):
                bad_file.sed_inplace("x", "y")



# @public -- populate __all__
#
# Copyright (C) 2016 Barry Warsaw <barry@python.org>
#
# This project is licensed under the terms of the Apache 2.0 License.

from typing import overload, TypeVar, Any
from collections.abc import Callable
import sys

__all__ = ('public', 'private')

ModuleAware = TypeVar("ModuleAware", bound=Callable[..., Any])


@overload
def public(thing: ModuleAware) -> ModuleAware:
    ...
@overload
def public(**kwargs: Any) -> Any | tuple[Any]:
    ...
def public(thing: Any | None = None, **kwargs: Any) -> ModuleAware | Any | tuple[Any]:
    """Add a name or names to __all__"""
    mdict = (sys._getframe(1).f_globals  # The function call syntax.
             if thing is None else
             sys.modules[thing.__module__].__dict__)  # The decorator syntax.
    dunder_all = mdict.setdefault("__all__", [])
    if not isinstance(dunder_all, list):
        raise TypeError(f"__all__ must be a list not: {type(dunder_all)}")
    if thing is None:
        # The function call form.
        retval = []
        for key, value in kwargs.items():
            if key not in dunder_all:
                dunder_all.append(key)
            mdict[key] = value
            retval.append(value)
        return retval[0] if len(retval) == 1 else tuple(retval)
    else:
        # The decorator form.
        if kwargs:
            raise AssertionError("Keyword arguments are incompatible with use "
                                 "as decorator")
        if thing.__name__ not in dunder_all:
            dunder_all.append(thing.__name__)
    return thing


def private(thing: ModuleAware) -> ModuleAware:
    """Remove names from __all__"""
    mdict = sys.modules[thing.__module__].__dict__
    dunder_all = mdict.setdefault("__all__", [])
    if not isinstance(dunder_all, list):
        raise TypeError(f"__all__ must be a list not: {type(dunder_all)}")
    if thing.__name__ in dunder_all:
        dunder_all.remove(thing.__name__)
    return thing


del ModuleAware
del overload, TypeVar, Callable



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest

import utlx
from utlx import Borg


class TestBorgPattern(unittest.TestCase):

    def test_shared_state_across_instances(self):
        a = Borg()
        b = Borg()
        a.x = 42
        self.assertEqual(b.x, 42)
        b.y = "hello"
        self.assertEqual(a.y, "hello")

    def test_dict_identity(self):
        a = Borg()
        b = Borg()
        self.assertIs(a.__dict__, b.__dict__)

    def test_modifying_state_affects_all_instances(self):
        a = Borg()
        b = Borg()
        a.data = [1, 2, 3]
        b.data.append(4)
        self.assertEqual(a.data, [1, 2, 3, 4])

    def test_new_instance_has_existing_state(self):
        Borg().shared = "yes"
        new_instance = Borg()
        self.assertEqual(new_instance.shared, "yes")

    def test_state_persistence_after_deletion(self):
        a = Borg()
        a.temp = "value"
        del a
        b = Borg()
        self.assertEqual(b.temp, "value")

    def test_independent_instances_same_state(self):
        class SubBorg(Borg):
            pass

        a = Borg()
        b = SubBorg()
        a.flag = True
        self.assertTrue(b.flag)
        self.assertIs(a.__dict__, b.__dict__)



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
import sys
import types
import tempfile
import importlib
from pathlib import Path
from configparser import ConfigParser, DEFAULTSECT

import utlx
from utlx import config


class TestConfigFunctions(unittest.TestCase):

    def setUp(self):
        # Create a temporary directory for config files
        self.temp_dir = tempfile.TemporaryDirectory()
        self.addCleanup(self.temp_dir.cleanup)
        self.temp_path = Path(self.temp_dir.name)

    def create_config_file(self, filename: str, section: str, **entries):
        """Helper to create a config file with given section and entries."""
        parser = ConfigParser()
        parser[section] = {k: str(v) for k, v in entries.items()}
        file_path = self.temp_path / filename
        with open(file_path, "w", encoding="utf-8") as f:
            parser.write(f)
        return file_path

    def test_get_config_with_section(self):
        cfg_file = self.create_config_file("test.ini", "mysection", key1="value1")
        section = config.get_config(cfg_file, "mysection")
        self.assertEqual(section["key1"], "value1")

    def test_get_config_without_section(self):
        # DEFAULT section
        parser = ConfigParser()
        parser["DEFAULT"] = {"key2": "value2"}
        cfg_file = self.temp_path / "default.ini"
        with open(cfg_file, "w", encoding="utf-8") as f:
            parser.write(f)
        section = config.get_config(cfg_file)
        self.assertEqual(section["key2"], "value2")

    def test_get_config_missing_file(self):
        # Should return an empty section if file does not exist
        section = config.get_config(self.temp_path / "nofile.ini", "somesect")
        self.assertEqual(len(section), 0)

    def test_get_config_empty_file(self):
        cfg_file = self.temp_path / "empty.ini"
        cfg_file.write_text("", encoding="utf-8")
        # Without section -> returns DEFAULT (empty)
        section = config.get_config(cfg_file)
        self.assertEqual(len(section), 0)

    def test_get_config_missing_section_in_file(self):
        cfg_file = self.create_config_file("other.ini", "othersect", foo="bar")
        # Asking for a section that does not exist -> returns DEFAULT (empty)
        section = config.get_config(cfg_file, "nonexistent")
        self.assertEqual(len(section), 0)

    def test_make_config_sets_globals(self):
        cfg_file = self.create_config_file("conf.ini", "pkgname", foo="bar")
        # Simulate a caller module's globals
        fake_globals = {
            "__file__": str(self.temp_path / "fake_module.py"),
            "__package__": "pkgname"
        }

        # Patch sys._getframe to return our fake globals
        class FakeFrame:
            def __init__(self, f_globals):
                self.f_globals = f_globals

        original_getframe = sys._getframe
        sys._getframe = lambda depth=0: FakeFrame(fake_globals)
        try:
            config.make_config("conf.ini")
        finally:
            sys._getframe = original_getframe

        self.assertIn("config", fake_globals)
        self.assertIn("set_config", fake_globals)
        self.assertEqual(fake_globals["config"]["foo"], "bar")
        self.assertTrue(callable(fake_globals["set_config"]))

    def test_set_config_updates_and_removes(self):
        # Prepare fake package and __config__ module
        package_name = "fakepkg"
        config_mod_name = package_name + ".__config__"

        fake_config_proxy = {"a": "1", "b": "2"}
        fake_config_module = types.SimpleNamespace(config=fake_config_proxy)
        fake_package_module = types.SimpleNamespace()

        sys.modules[package_name] = fake_package_module
        sys.modules[config_mod_name] = fake_config_module
        # Add a dummy submodule to be removed
        submodule_name = package_name + ".submod"
        sys.modules[submodule_name] = types.SimpleNamespace()

        fglobals = {"__package__": package_name}

        # Patch importlib.reload to mark reload called
        reloaded = {}
        original_reload = importlib.reload
        importlib.reload = lambda mod: reloaded.setdefault("called", True) or mod
        try:
            config.set_config(fglobals, a="10", b=None, c="3")
        finally:
            importlib.reload = original_reload

        # 'a' updated, 'b' removed, 'c' added
        self.assertEqual(fake_config_proxy["a"], "10")
        self.assertNotIn("b", fake_config_proxy)
        self.assertEqual(fake_config_proxy["c"], "3")
        # Submodule should be removed
        self.assertNotIn(submodule_name, sys.modules)
        # Reload should have been called
        self.assertTrue(reloaded.get("called"))

    def test_set_config_no_changes(self):
        """Calling set_config with no changes should not break anything."""
        package_name = "fakepkg2"
        config_mod_name = package_name + ".__config__"

        fake_config_proxy = {"x": "1"}
        fake_config_module = types.SimpleNamespace(config=fake_config_proxy)
        fake_package_module = types.SimpleNamespace()

        sys.modules[package_name] = fake_package_module
        sys.modules[config_mod_name] = fake_config_module

        fglobals = {"__package__": package_name}

        original_reload = importlib.reload
        importlib.reload = lambda mod: mod
        try:
            config.set_config(fglobals)
        finally:
            importlib.reload = original_reload

        self.assertEqual(fake_config_proxy["x"], "1")  # unchanged



# flake8-in-file-ignores: noqa: F401,F403,F821

# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

from .__about__ import * ; del __about__  # type: ignore[name-defined]

from ._defined       import * ; del _defined        # type: ignore[name-defined]
from ._deprecated    import * ; del _deprecated     # type: ignore[name-defined]
from ._public        import * ; del _public         # type: ignore[name-defined]
from ._classproperty import * ; del _classproperty  # type: ignore[name-defined]
from ._rwproperty    import * ; del _rwproperty     # type: ignore[name-defined]
from ._cache         import * ; del _cache          # type: ignore[name-defined]
from ._renumerate    import * ; del _renumerate     # type: ignore[name-defined]
from ._adict         import * ; del _adict          # type: ignore[name-defined]
from ._borg          import * ; del _borg           # type: ignore[name-defined]
from ._const         import * ; del _const          # type: ignore[name-defined]
from ._modpath       import * ; del _modpath        # type: ignore[name-defined]
from ._dllpath       import * ; del _dllpath        # type: ignore[name-defined]
from ._ftee          import * ; del _ftee           # type: ignore[name-defined]
from ._run           import * ; del _run            # type: ignore[name-defined]
from ._unique        import * ; del _unique         # type: ignore[name-defined]
from ._util          import * ; del _util           # type: ignore[name-defined]
from .epath          import * ; del epath
from . import platform
from . import ctypes
from . import config
from . import imports
from . import epath



# Copyright (c) 2012 Adam Karpierz
# SPDX-License-Identifier: Zlib

from typing import TYPE_CHECKING, TypeAlias, Any
from collections.abc import Iterable
import subprocess
import logging

__all__ = ('run',)

log = logging.getLogger(__name__)


class run:
    """Unified subprocess interface with namespaced constants and callable execution."""

    if TYPE_CHECKING:
        CompletedProcess:      TypeAlias = subprocess.CompletedProcess[str | bytes]
        CompletedTextProcess:  TypeAlias = subprocess.CompletedProcess[str]
        CompletedBytesProcess: TypeAlias = subprocess.CompletedProcess[bytes]
    else:
        CompletedProcess      = subprocess.CompletedProcess
        CompletedTextProcess  = subprocess.CompletedProcess
        CompletedBytesProcess = subprocess.CompletedProcess

    PIPE    = subprocess.PIPE
    STDOUT  = subprocess.STDOUT
    DEVNULL = subprocess.DEVNULL

    SubprocessError    = subprocess.SubprocessError
    TimeoutExpired     = subprocess.TimeoutExpired
    CalledProcessError = subprocess.CalledProcessError

    class SafeString(str):
        """Marks sensitive arguments to be masked in logs."""

    def __new__(cls, *args: Any,  # type: ignore[misc]
                start_terminal_window: bool = False, **kwargs: Any) -> CompletedProcess:
        """Runs the command described by `args`.

        Waits for the command to complete and returns a run.CompletedProcess instance.

        Args:
            args: Command-line arguments passed to `subprocess.run`.
            start_terminal_window: If True, starts the command in a separate console
                                   window (server mode).
            kwargs: Additional keyword arguments controlling subprocess execution.

        Returns:
            run.CompletedProcess: Result of the executed command.
        """
        if start_terminal_window:  # pragma: no cover
            args = ("cmd.exe", "/C", "start", *args)
        output: run.CompletedProcess = subprocess.run(
            [str(arg) for arg in args], check=kwargs.pop("check", True), **kwargs)
        print_cmd = [("*****" if isinstance(arg, cls.SafeString) else arg) for arg in args]
        log.debug("cmd=%s, returncode=%d", str(print_cmd), int(output.returncode))
        return output

    @staticmethod
    def split_kwargs(kwargs: dict[str, Any], forbidden_kwargs: Iterable[str]) \
            -> tuple[dict[str, Any], dict[str, Any]]:
        allowed_kwargs  = {key: val for key, val in kwargs.items()
                           if key not in forbidden_kwargs}
        reserved_kwargs = {key: val for key, val in kwargs.items()
                           if key in forbidden_kwargs}
        return (allowed_kwargs, reserved_kwargs)



# flake8-in-file-ignores: noqa: E305,F401,F821

# Copyright (c) 1994 Adam Karpierz
# SPDX-License-Identifier: Zlib

from __future__ import annotations

import ctypes as ct
from ctypes import CDLL as DLL
try:
    from _ctypes import dlclose  # type: ignore[attr-defined,unused-ignore]
except ImportError:  # pragma: no cover
    dlclose = lambda handle: None
from ctypes import CFUNCTYPE as CFUNC

from ... import ctypes as ctx

__all__ = (
    'DLL', 'dlclose', 'CFUNC',
    'time_t', 'suseconds_t', 'timeval',
    'SOCKET', 'INVALID_SOCKET', 'socklen_t', 'sa_family_t', 'in_addr_t', 'in_port_t',
    'sockaddr', 'in_addr', 'sockaddr_in', 'in6_addr', 'sockaddr_in6',
    'FD_SETSIZE', 'fd_set', 'FD_ZERO', 'FD_ISSET', 'FD_SET', 'FD_CLR', 'select',
)

# X32 kernel interface is 64-bit.
if False:  # if defined __x86_64__ && defined __ILP32__
    # quad_t is also 64 bits.
    time_t = suseconds_t = ct.c_longlong
else:
    time_t = suseconds_t = ct.c_long
# endif

# Taken from the file <sys/time.h>
# #include <time.h>
#
# struct timeval {
#     time_t      tv_sec;   /* Seconds. */
#     suseconds_t tv_usec;  /* Microseconds. */
# };

class timeval(ct.Structure):
    _fields_ = [
    ("tv_sec",  time_t),       # seconds
    ("tv_usec", suseconds_t),  # microseconds
]

# Taken from the file libpcap's "socket.h"

# Some minor differences between sockets on various platforms.
# We include whatever sockets are needed for Internet-protocol
# socket access.

# In UN*X, a socket handle is a file descriptor, and therefore
# a signed integer.
SOCKET = ct.c_int

# In UN*X, the error return if socket() fails is -1.
INVALID_SOCKET = SOCKET(-1).value

class sockaddr(ct.Structure):
    _fields_ = [
    ("sa_family", ct.c_short),
    ("__pad1",    ct.c_ushort),
    ("ipv4_addr", ct.c_byte * 4),
    ("ipv6_addr", ct.c_byte * 16),
    ("__pad2",    ct.c_ulong),
]

# https://yarchive.net/comp/linux/socklen_t.html
socklen_t = ct.c_uint32 if ct.sizeof(ct.c_uint) < 4 else ct.c_uint

# POSIX.1g specifies this type name for the `sa_family' member.
sa_family_t = ct.c_ushort

# Types to represent an address and port.
in_addr_t = ct.c_uint32
in_port_t = ct.c_uint16

# IPv4 AF_INET sockets:

class in_addr(ct.Structure):
    _fields_ = [
    ("s_addr", in_addr_t),
]

class sockaddr_in(ct.Structure):
    _fields_ = [
    ("sin_family", sa_family_t),      # e.g. AF_INET, AF_INET6
    ("sin_port",   in_port_t),        # Port number.
    ("sin_addr",   in_addr),          # Internet address.
    ("sin_zero",   (ct.c_ubyte        # Pad to size of `struct sockaddr'.
                    * (ct.sizeof(sockaddr)
                       - ct.sizeof(sa_family_t)
                       - ct.sizeof(in_port_t)
                       - ct.sizeof(in_addr)))),
]

# IPv6 AF_INET6 sockets:

class in6_addr(ct.Union):
    _fields_ = [
    ("s6_addr",   (ct.c_uint8 * 16)),
    ("s6_addr16", (ct.c_uint16 * 8)),
    ("s6_addr32", (ct.c_uint32 * 4)),
]

class sockaddr_in6(ct.Structure):
    _fields_ = [
    ("sin6_family",   sa_family_t),   # address family, AF_INET6
    ("sin6_port",     in_port_t),     # Transport layer port #
    ("sin6_flowinfo", ct.c_uint32),   # IPv6 flow information
    ("sin6_addr",     in6_addr),      # IPv6 address
    ("sin6_scope_id", ct.c_uint32),   # IPv6 scope-id
]

# From <sys/select.h>

# The fd_set member
_fd_mask = ct.c_long  # fd_mask and NFDBITS are not portable, cannot be exported.
_NFDBITS = 8 * ct.sizeof(_fd_mask)

# Maximum number of file descriptors in `fd_set'.
FD_SETSIZE = 1024

class fd_set(ct.Structure):
    _fields_ = [
    ("fds_bits", (_fd_mask * (FD_SETSIZE // _NFDBITS))),
]

def FD_ZERO(fdsetp: ctx.POINTER[fd_set]) -> None:
    import ctypes as ct
    fdset = fdsetp.contents
    ct.memset(fdsetp, 0, ct.sizeof(fdset))
FD_ZERO = CFUNC(None, ct.POINTER(fd_set))(FD_ZERO)

def FD_ISSET(fd: int, fdsetp: ctx.POINTER[fd_set]) -> int:
    fdset = fdsetp.contents
    return int(fdset.fds_bits[fd // _NFDBITS] & (1 << (fd % _NFDBITS)))
FD_ISSET = CFUNC(ct.c_int, ct.c_int, ct.POINTER(fd_set))(FD_ISSET)

def FD_SET(fd: int, fdsetp: ctx.POINTER[fd_set]) -> None:
    fdset = fdsetp.contents
    fdset.fds_bits[fd // _NFDBITS] |= (1 << (fd % _NFDBITS))
FD_SET = CFUNC(None, ct.c_int, ct.POINTER(fd_set))(FD_SET)

def FD_CLR(fd: int, fdsetp: ctx.POINTER[fd_set]) -> None:
    fdset = fdsetp.contents
    fdset.fds_bits[fd // _NFDBITS] &= ~(1 << (fd % _NFDBITS))
FD_CLR = CFUNC(None, ct.c_int, ct.POINTER(fd_set))(FD_CLR)

libc = ct.CDLL("libSystem.dylib")

select = libc.select
select.restype  = ct.c_int
select.argtypes = [ct.c_int,
                   ct.POINTER(fd_set), ct.POINTER(fd_set), ct.POINTER(fd_set),
                   ct.POINTER(timeval)]

del ct, ctx, libc



# Copyright (c) 2016 Adam Karpierz
# SPDX-License-Identifier: Zlib

from typing import TypeAlias, Any
from typing_extensions import Self
from collections.abc import Callable, Iterable, Generator
from os import PathLike
import sys
import os
import re
import stat
import shutil
import tempfile
import pathlib
import hashlib
import contextlib

import charset_normalizer

__all__ = ('Path',)

StrPath:     TypeAlias = str | PathLike[str]
AnyCallable: TypeAlias = Callable[..., Any]

_HAS_FILE_ATTRS = hasattr(os.stat_result, "st_file_attributes")


class Path(pathlib.Path):

    if sys.version_info[:2] <= (3, 12):  # pragma: no cover
        """Constructor"""
        def __new__(cls, *args: Any, **kwargs: Any) -> Self:
            cls._flavour = (pathlib.WindowsPath  # type: ignore[union-attr]
                            if os.name == "nt" else
                            pathlib.PosixPath)._flavour
            return super().__new__(cls, *args, **kwargs)

    if sys.version_info[:2] <= (3, 11):  # pragma: no cover
        def is_relative_to(self, other: StrPath) -> bool:  # type: ignore[override]
            return super().is_relative_to(other)

        def relative_to(self, other: StrPath) -> Self:  # type: ignore[override]
            return super().relative_to(other)

    if sys.version_info[:2] <= (3, 9):  # pragma: no cover
        def hardlink_to(self, target: StrPath) -> None:
            Path(target).link_to(self)

    def exists(self) -> bool:
        return super().exists() or self._is_real_link()

    def mkdir(self, mode: int = 0o777,
              parents: bool = False, exist_ok: bool = True) -> None:
        return super().mkdir(mode=mode, parents=parents, exist_ok=exist_ok)

    def rmdir(self, *, ignore_errors: bool = False,
              onexc: Callable[[AnyCallable, str, Any],
                              object] | None = None) -> None:
        if not self.exists():
            return
        shutil.rmtree(self, ignore_errors=ignore_errors,
                      onerror=onexc or self.__remove_readonly)

    @staticmethod
    def __remove_readonly(func: AnyCallable, path: str, excinfo: Any) -> None:
        os.chmod(path, stat.S_IWRITE)
        func(path)

    def cleardir(self, *, ignore_errors: bool = False,
                 onexc: Callable[[AnyCallable, str, Any],
                                 object] | None = None) -> None:
        if not self.exists():
            return
        if not self.is_dir():
            raise NotADirectoryError(f"The directory name is invalid: '{self}'")
        if self._is_real_link():
            raise NotADirectoryError("Cannot call cleardir on a symbolic link")
        for entry in self.iterdir():
            if entry.is_dir() and not entry.is_symlink():
                entry.rmdir(ignore_errors=ignore_errors, onexc=onexc)
            else:
                entry.unlink(missing_ok=True)

    def _is_real_link(self) -> bool:
        if _HAS_FILE_ATTRS:
            # Special handling for directory junctions to make them behave like
            # symlinks for shutil.rmtree, since in general they do not appear as
            # regular links.
            try:
                st = os.lstat(self)
                return bool(stat.S_ISLNK(st.st_mode)
                            or (st.st_file_attributes & stat.FILE_ATTRIBUTE_REPARSE_POINT
                                and (not hasattr(os.stat_result, "st_reparse_tag")
                                     or st.st_reparse_tag == stat.IO_REPARSE_TAG_MOUNT_POINT)))
            except OSError:
                return False
        else:
            return os.path.islink(self)

    def copydir(self, dst: StrPath, *, symlinks: bool = False,
                ignore: Callable[[str, list[str]], Iterable[str]] | None = None,
                copy_function: Callable[[str, str], object] | None = None,
                ignore_dangling_symlinks: bool = False,
                dirs_exist_ok: bool = False) -> Self:
        return type(self)(shutil.copytree(self, dst, symlinks=symlinks, ignore=ignore,
                                          copy_function=copy_function or shutil.copy2,
                                          ignore_dangling_symlinks=ignore_dangling_symlinks,
                                          dirs_exist_ok=dirs_exist_ok))

    def unlink(self, missing_ok: bool = True) -> None:
        try:
            return super().unlink(missing_ok=missing_ok)
        except PermissionError:
            self.chmod(stat.S_IWRITE)
            return super().unlink(missing_ok=missing_ok)

    def copy(self, dst: StrPath, *, follow_symlinks: bool = True) -> Self:
        return type(self)(shutil.copy2(self, dst, follow_symlinks=follow_symlinks))

    def move(self, dst: StrPath, *,
             copy_function: Callable[[str, str], object] | None = None) -> Self | None:
        if not self.exists():
            return None
        return type(self)(shutil.move(self, dst, copy_function=copy_function or shutil.copy2))

    def copystat(self, dst: StrPath, *, follow_symlinks: bool = True) -> None:
        return shutil.copystat(self, dst, follow_symlinks=follow_symlinks)

    @classmethod
    def which(cls, cmd: StrPath, *, mode: int = os.F_OK | os.X_OK,
              path: StrPath | None = None) -> Self | None:
        result = shutil.which(str(cmd), mode=mode, path=path)
        return cls(result) if result is not None else None

    def file_hash(self, algorithm: str, *, chuck_size: int = 65536) -> Any:
        constructor = self.__hash_algorithms.get(algorithm, lambda: hashlib.new(algorithm))
        hash_value = constructor()
        with self.open("rb") as f:
            while True:
                chunk = f.read(chuck_size)
                if not chunk: break
                hash_value.update(chunk)
        return hash_value

    def dir_hash(self, algorithm: str, *, chuck_size: int = 65536) -> Any:
        constructor = self.__hash_algorithms.get(algorithm, lambda: hashlib.new(algorithm))
        hash_value = constructor()
        for root, dirs, files in os.walk(self):
            for name in files:
                fpath = Path(root)/name
                with fpath.open("rb") as f:
                    while True:
                        chunk = f.read(chuck_size)
                        if not chunk: break
                        hash_value.update(chunk)
        return hash_value

    __hash_algorithms: dict[str, Callable[[], Any]] = {
        "md5":     hashlib.md5,
        "sha1":    hashlib.sha1,
        "sha224":  hashlib.sha224,
        "sha256":  hashlib.sha256,
        "sha384":  hashlib.sha384,
        "sha512":  hashlib.sha512,
        "blake2b": hashlib.blake2b,
        "blake2s": hashlib.blake2s,
    }

    def unpack_archive(self, extract_dir: StrPath | None = None, *,
                       format: str | None = None) -> None:  # noqa: A002
        """Unpack an archive."""
        return shutil.unpack_archive(self, extract_dir, format)

    def sed_inplace(self, pattern: str | re.Pattern[str], repl: str, *,
                    flags: int | re.RegexFlag = 0, encoding: str | None = None) -> None:
        """Perform the pure-Python equivalent of in-place `sed` substitution: e.g., \
        `sed -i -e 's/'${pattern}'/'${repl}'/g "${filename}"`."""

        # For efficiency, precompile the passed regular expression.
        if not isinstance(pattern, re.Pattern): pattern = re.compile(pattern, flags)

        if encoding is not None:
            content = self.open("rt", encoding=encoding, newline="").read()
        else:
            data = self.read_bytes()
            detected = charset_normalizer.from_bytes(data).best()
            try:
                if detected:
                    encoding = detected.encoding
                    content  = str(detected)
                else:
                    encoding = None
                    content  = data.decode()
            except Exception:
                raise UnicodeError(f"The file '{self}' cannot be decoded. "
                                   f"It appears to be a binary file.")

        with tempfile.NamedTemporaryFile(mode="wt", encoding=encoding,
                                         newline="", delete=False) as tmp_file:
            if flags & re.MULTILINE:
                tmp_file.write(pattern.sub(repl, content))
            else:
                for line in content.splitlines(keepends=True):
                    tmp_file.write(pattern.sub(repl, line))
        # Overwrite the original file with the munged temporary file
        # in a manner preserving file attributes (e.g., permissions).
        shutil.copystat(self, tmp_file.name)
        shutil.move(tmp_file.name, self)

    def chdir(self) -> None:
        os.chdir(self)

    @contextlib.contextmanager
    def pushd(self) -> Generator[None, None, None]:
        curr_dir = os.getcwd()
        os.chdir(self)
        try:
            yield
        finally:
            os.chdir(curr_dir)


del Self, Callable, Iterable, Generator, PathLike
del StrPath, AnyCallable



# Copyright (c) 1994 Adam Karpierz
# SPDX-License-Identifier: Zlib

import sys
import os
import platform

__all__ = (
    'is_windows', 'is_wsl', 'is_cygwin', 'is_msys', 'is_linux', 'is_macos',
    'is_bsd', 'is_sunos', 'is_aix', 'is_android', 'is_posix', 'is_32bits',
    'is_ucs2', 'is_cpython', 'is_pypy', 'is_ironpython',
)

is_windows = (bool(platform.win32_ver()[0])
              or (sys.platform in ("win32", "cygwin", "msys"))
              or (sys.platform == "cli" and os.name in ("nt", "ce"))
              or (os.name == "java"
                  and "windows" in platform.java_ver()[3][0].lower()))
is_wsl     = ("microsoft-standard" in platform.uname().release)
is_cygwin  = (sys.platform == "cygwin")
is_msys    = (sys.platform == "msys")
is_linux   = sys.platform.startswith("linux")
is_macos   = sys.platform.startswith("darwin")
is_bsd     = sys.platform.startswith(("freebsd", "openbsd", "netbsd"))
is_sunos   = sys.platform.startswith(("sunos", "solaris"))
is_aix     = sys.platform.startswith("aix")
is_android = hasattr(sys, "getandroidapilevel")
is_posix   = (os.name == "posix")
is_32bits  = (sys.maxsize <= 2**32)
is_ucs2    = (sys.maxunicode < 0x10FFFF)
is_cpython = (platform.python_implementation().lower() == "cpython")
is_pypy    = (platform.python_implementation().lower() == "pypy")
is_ironpython = (platform.python_implementation().lower() == "ironpython"
                 or "cli" in (platform.system().lower(), sys.platform))

del sys, os, platform



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest

import utlx
from utlx import classproperty


class TestClassProperty(unittest.TestCase):

    def test_basic_classproperty_access(self):
        class MyClass:
            @classproperty
            def name(cls):
                return cls.__name__

        self.assertEqual(MyClass.name, "MyClass")

    def test_classproperty_access_from_instance(self):
        class MyClass:
            @classproperty
            def info(cls):
                return "class-level"

        instance = MyClass()
        self.assertEqual(instance.info, "class-level")

    def test_classproperty_with_docstring(self):
        class MyClass:
            @classproperty
            def value(cls):
                """Returns a fixed value"""
                return 42

        self.assertEqual(MyClass.value, 42)
        self.assertEqual(MyClass.__dict__['value'].__doc__, "Returns a fixed value")

    def test_classproperty_without_fget_raises(self):
        cp = classproperty()
        with self.assertRaises(AttributeError):
            cp.__get__(None, None)

    def test_classproperty_with_non_callable_fget_raises(self):
        cp = classproperty(fget=123)
        with self.assertRaises(TypeError):
            cp.__get__(None, None)

    def test_classproperty_rejects_fset_and_fdel(self):
        with self.assertRaises(ValueError):
            classproperty(lambda cls: 1, fset=lambda cls, val: None)

        with self.assertRaises(ValueError):
            classproperty(lambda cls: 1, fdel=lambda cls: None)

    def test_classproperty_getter_method(self):
        def get_value(cls):
            return "OK"

        cp = classproperty().getter(get_value)
        class Dummy:
            prop = cp

        self.assertEqual(Dummy.prop, "OK")

    def test_classproperty_doc_inferred(self):
        def get(cls):
            """Auto-doc"""
            return "doc"  # pragma: no cover

        cp = classproperty(get)
        self.assertEqual(cp.__doc__, "Auto-doc")

    def test_classproperty_fget_property(self):
        def get(cls):
            return "value"  # pragma: no cover

        cp = classproperty(get)
        self.assertEqual(cp.fget, get)



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest

import utlx
from utlx import cached, cached_property


class TestCachedDecorators(unittest.TestCase):

    def test_cached_method_caches_result(self):
        class Example:
            def __init__(self):
                self.counter = 0

            @cached
            def compute(self):
                self.counter += 1
                return self.counter

        obj = Example()
        result1 = obj.compute()
        result2 = obj.compute()
        self.assertEqual(result1, result2)
        self.assertEqual(obj.counter, 1)

    def test_cached_method_initializes_cache(self):
        class Example:
            @cached
            def value(self):
                return 42

        obj = Example()
        self.assertFalse(hasattr(obj, '__cache__'))
        val = obj.value()
        self.assertTrue(hasattr(obj, '__cache__'))
        self.assertEqual(val, 42)

    def test_cached_property_basic_usage(self):
        class Example:
            def __init__(self):
                self.counter = 0

            @cached_property
            def prop(self):
                self.counter += 1
                return self.counter

        obj = Example()
        val1 = obj.prop
        val2 = obj.prop
        self.assertEqual(val1, val2)
        self.assertEqual(obj.counter, 1)

    def test_cached_property_with_setter_and_deleter(self):
        class Example:
            def __init__(self):
                self._value = 10

            def get(self):
                return self._value

            def set(self, val):
                self._value = val

            def delete(self):
                self._value = None

            prop = cached_property(get, set, delete, doc="cached prop")

        obj = Example()
        self.assertEqual(obj.prop, 10)
        obj.prop = 20
        self.assertEqual(obj.prop, 20)
        del obj.prop
        self.assertIsNone(obj._value)

    def test_cached_property_with_setter_and_deleter_first_set(self):
        class Example:
            def __init__(self):
                self._value = 10

            def get(self):
                return self._value

            def set(self, val):
                self._value = val

            def delete(self):
                self._value = None

            prop = cached_property(get, set, delete, doc="cached prop")

        obj = Example()
        obj.prop = 20
        self.assertEqual(obj.prop, 20)
        del obj.prop
        self.assertIsNone(obj._value)

    def test_cached_property_with_setter_and_deleter_first_delete(self):
        class Example:
            def __init__(self):
                self._value = 10

            def get(self):
                return self._value

            def set(self, val):
                self._value = val

            def delete(self):
                self._value = None

            prop = cached_property(get, set, delete, doc="cached prop")

        obj = Example()
        del obj.prop
        self.assertIsNone(obj._value)
        obj.prop = 20
        self.assertEqual(obj.prop, 20)

    def test_cached_property_without_getter(self):
        class Example:
            def __init__(self):
                self._value = 10

            def set(self, val):
                self._value = val

            def delete(self):
                self._value = None

            prop = cached_property(None, set, delete, doc="cached prop")

        obj = Example()
        self.assertEqual(obj._value, 10)
        obj.prop = 20
        self.assertEqual(obj._value, 20)
        del obj.prop
        self.assertIsNone(obj._value)

    def test_cached_property_docstring(self):
        class Example:
            @cached_property
            def prop(self):
                """Returns something"""
                return 1

        self.assertEqual(Example.__dict__['prop'].__doc__, "Returns something")
        self.assertEqual(Example().prop, 1)  # for full coverage



# Copyright (c) 1994 Adam Karpierz
# SPDX-License-Identifier: Zlib

__all__ = ('arch',)


def get_python_arch() -> str | None:
    import sys
    import platform
    machine = platform.machine().lower()
    is_32bits = (sys.maxsize <= 2**32)
    little_endian = sys.byteorder.lower() == "little"

    X86_64      = ("x86_64", "amd64")
    X86_32      = ("i386", "i486", "i586", "i686", "x86", "ia32")
    ARM_64      = ("aarch64", "arm64")
    ARM_64_BE   = ("aarch64_be", "arm64_be")
    ARM_32_SOFT = ("armv6l", "armv5tel", "armv4tl", "armv4l", "arml", "armel")
    ARM_32_HARD = ("armv8l", "armv7l")
    PPC_64_LE   = ("ppc64le", "powerpc64le")
    PPC_64      = ("ppc64", "powerpc64")
    PPC_32_LE   = ("ppcle", "powerpcle")
    PPC_32      = ("ppc", "powerpc")
    MIPS_64_LE  = ("mips64el", "mips64le")
    MIPS_64_BE  = ("mips64", "mips64eb")
    MIPS_32_LE  = ("mipsel", "mipsle")
    MIPS_32_BE  = ("mips", "mipseb")
    RISC_V_64   = ("riscv64",)
    S390X       = ("s390x",)

    # x86
    if machine in X86_64:
        arch = "x86_64" if not is_32bits else "x86"
    elif machine in X86_32:
        arch = "x86"
    # ARM 64-bit
    elif machine in ARM_64 and not is_32bits:
        suffix = "" if little_endian else "be"
        arch = "aarch64" + suffix
    elif machine in ARM_64_BE and not is_32bits:
        suffix = "be"
        arch = "aarch64" + suffix
    # ARM 32-bit
    elif machine in ARM_32_SOFT:
        suffix = "le"
        arch = "arm" + suffix
    elif (machine in ARM_32_HARD or (machine in ARM_64 and is_32bits)):
        # hf = hard-float,little-endian, le = soft-float,little-endian
        suffix = "hf" if has_fpu() else "le"
        arch = "arm" + suffix
    elif ((machine.startswith("arm") and machine.endswith("b"))
          or (machine in ARM_64_BE and is_32bits)):
        arch = None  # 32-bit big endian is not supported
    # POWER / ppc
    elif machine in PPC_64_LE or machine in PPC_64:
        suffix = "le" if little_endian else ""
        arch = ("ppc64" if not is_32bits else "ppc") + suffix
    elif machine in PPC_32_LE or machine in PPC_32:
        suffix = "le" if little_endian else ""
        arch = "ppc" + suffix
    # MIPS
    elif machine in MIPS_64_LE or machine in MIPS_64_BE:
        suffix = "le" if little_endian else ""
        arch = ("mips64" if not is_32bits else "mips") + suffix
    elif machine in MIPS_32_LE or machine in MIPS_32_BE:
        suffix = "le" if little_endian else ""
        arch = "mips" + suffix
    # RISC-V
    elif machine in RISC_V_64:
        arch = "riscv64"
    # IBM Z
    elif machine in S390X:
        arch = "s390x"
    else:
        arch = None  # unknown or unsupported

    return arch


def has_fpu() -> bool:
    """
    Detects presence of FPU on ARM:

    1. Try /proc/cpuinfo
    2. Fallback: readelf on Python interpreter
    """
    import sys
    import re
    import subprocess
    try:
        with open("/proc/cpuinfo") as f:
            for line in f:
                line = line.lower()
                if "vfp" in line or "neon" in line:
                    return True
    except FileNotFoundError:
        pass
    try:
        out = subprocess.check_output(["readelf", "-A", sys.executable],
                                      stderr=subprocess.DEVNULL, text=True)
        if re.search(r"vfp|neon", out, re.IGNORECASE):
            return True
    except (OSError, subprocess.CalledProcessError):
        pass
    return False


arch = get_python_arch()



# Copyright (c) 1994 Adam Karpierz
# SPDX-License-Identifier: Zlib

__all__ = ('arch',)


def get_python_arch() -> str | None:
    import sys
    import platform
    machine = platform.machine().lower()
    is_32bits = (sys.maxsize <= 2**32)

    # x86
    if machine in ("amd64", "x86_64"):
        arch = "x86_64" if not is_32bits else "x86"
    elif machine in ("x86", "i386", "i686", "ia32"):
        arch = "x86"
    # ARM
    elif machine in ("aarch64",) or machine.startswith("arm"):
        arch = "arm64" if not is_32bits else "arm"
    else:
        arch = None  # unknown or unsupported

    return arch


arch = get_python_arch()



# Copyright (c) 2018 Adam Karpierz
# SPDX-License-Identifier: Zlib

from typing import TypeAlias, TextIO
from collections.abc import Generator
from os import PathLike
import sys
import io
import contextlib

__all__ = ('ftee',)

StrPath: TypeAlias = str | PathLike[str]


class _Tee(io.TextIOBase):

    def __init__(self, *files: TextIO) -> None:
        """Initializer"""
        self._files = files

    def close(self) -> None:
        super().close()
        self._files[0].flush()
        for f in self._files[1:]:
            f.close()

    def writable(self) -> bool:
        if self.closed:
            raise ValueError("I/O operation on closed file.")
        return True

    def write(self, s: str) -> int:
        count: int = 0
        for f in self._files:
            count = f.write(s)
        return count

    def flush(self) -> None:
        for f in self._files:
            f.flush()


@contextlib.contextmanager
def ftee(*filenames: StrPath) -> Generator[TextIO | _Tee, None, None]:
    stdout = sys.stdout
    files = [open(fname, "w") for fname in filenames]
    try:
        sys.stdout = _Tee(stdout, *files)
        yield sys.stdout
    finally:
        sys.stdout.close()
        sys.stdout = stdout


del TextIO, Generator, StrPath, PathLike
del io, contextlib



# Copyright (c) 2012 Adam Karpierz
# SPDX-License-Identifier: Zlib

import sys

__all__ = ('defined',)


def defined(varname: str,  # type: ignore[no-untyped-def]
            __getframe=sys._getframe) -> bool:
    frame = __getframe(1)
    return varname in frame.f_locals or varname in frame.f_globals


del sys



# Copyright (c) 2016 Adam Karpierz
# SPDX-License-Identifier: Zlib

from typing  import Any
from pathlib import Path
import configparser

__all__ = ('make_config',)


def make_config(cfg_fname: str, cfg_section: str | None = None) -> None:
    import sys
    from pathlib import Path
    from functools import partial
    fglobals = sys._getframe(1).f_globals
    fglobals.pop("__builtins__", None)
    fglobals.pop("__cached__",   None)
    if cfg_section is None: cfg_section = fglobals["__package__"]
    cfg_path = Path(fglobals["__file__"]).parent/cfg_fname
    fglobals["__all__"] = ("config", "set_config")
    fglobals["config"] = get_config(cfg_path, cfg_section)
    fglobals["set_config"] = partial(set_config, fglobals)


def get_config(cfg_path: Path | str,
               cfg_section: str | None = None) -> configparser.SectionProxy:
    from pathlib import Path
    from configparser import ConfigParser, ExtendedInterpolation, DEFAULTSECT
    cfg_path = Path(cfg_path)
    interpolation = ExtendedInterpolation()
    inline_comment_prefixes = ("#", ";")
    if cfg_section is not None:
        cfg = ConfigParser(interpolation=interpolation,
                           inline_comment_prefixes=inline_comment_prefixes,
                           default_section=cfg_section)
    else:
        cfg = ConfigParser(interpolation=interpolation,
                           inline_comment_prefixes=inline_comment_prefixes)
    cfg.read(cfg_path, encoding="utf-8")
    return cfg[cfg_section if cfg_section is not None else DEFAULTSECT]


def set_config(fglobals: dict[str, Any], **cfg_dict: Any) -> None:
    import sys
    import importlib
    # Update config
    to_update = {key: str(val) for key, val in cfg_dict.items()
                 if val is not None}
    to_remove = {key for key, val in cfg_dict.items() if val is None}
    package_name = fglobals["__package__"]
    config_name  = package_name + ".__config__"
    config = sys.modules[config_name].config
    config.update(to_update)
    for key in to_remove: config.pop(key, None)
    # Reload
    for mod_name in tuple(sys.modules):
        if (mod_name.startswith(package_name + ".")
           and mod_name != config_name):
            del sys.modules[mod_name]
    importlib.reload(sys.modules[package_name])


del Any
del Path
del configparser



# Copyright (c) 2012 Adam Karpierz
# SPDX-License-Identifier: Zlib

from typing import TypeVar, TypeAlias, Any
from collections.abc import Callable

__all__ = ('cached', 'cached_property')

_P = TypeVar("_P", bound=object)
_T = TypeVar("_T")

AnyCallable: TypeAlias = Callable[..., Any]


def cached(method: AnyCallable) -> AnyCallable:
    """Decorator to simple cache method's result"""

    from functools import wraps

    @wraps(method)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        key: int = hash(method)
        try:
            cache = self.__cache__
        except AttributeError:
            cache = self.__cache__ = {}
        try:
            result = cache[key]
        except KeyError:
            result = cache[key] = method(self, *args, **kwargs)
        return result

    return wrapper


def cached_property(fget: Callable[[_P], _T] | None = None,
                    fset: Callable[[_P, _T], None] | None = None,
                    fdel: Callable[[_P], None] | None = None,
                    doc: str | None = None) -> property:
    """Decorator to simple cache property attribute.

    fget
      function to be used for cached getting an attribute value
    fset
      function to be used for setting an attribute value
    fdel
      function to be used for deleting an attribute
    doc
      docstring
    """
    from functools import wraps

    _fget, _fset, _fdel = fget, fset, fdel
    if fget is not None:
        key: int = hash(fget)

        @wraps(fget)
        def _fget(self: _P) -> _T:  # noqa: F811
            try:
                cache = self.__cache__  # type: ignore[attr-defined]
            except AttributeError:
                cache = self.__cache__ = {}  # type: ignore[attr-defined]
            result: _T
            try:
                result = cache[key]
            except KeyError:
                result = cache[key] = fget(self)
            return result

        if fset is not None:
            @wraps(fset)
            def _fset(self: _P, value: _T) -> None:  # noqa: F811
                try:
                    del self.__cache__[key]  # type: ignore[attr-defined]
                except (AttributeError, KeyError):
                    pass
                fset(self, value)

        if fdel is not None:
            @wraps(fdel)
            def _fdel(self: _P) -> None:  # noqa: F811
                try:
                    del self.__cache__[key]  # type: ignore[attr-defined]
                except (AttributeError, KeyError):
                    pass
                fdel(self)

    return property(_fget, _fset, _fdel, doc)


del TypeVar, TypeAlias, Callable, AnyCallable



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
import sys
import tempfile
from pathlib import Path
from io import StringIO

import utlx
from utlx import ftee
from utlx._ftee import _Tee


class TestTee(unittest.TestCase):

    def test_write_writes_to_all_files(self):
        buf1 = StringIO()
        buf2 = StringIO()
        tee = _Tee(buf1, buf2)
        tee.write("hello")
        self.assertEqual(buf1.getvalue(), "hello")
        self.assertEqual(buf2.getvalue(), "hello")

    def test_flush_flushes_all_files(self):
        buf1 = StringIO()
        buf2 = StringIO()
        tee = _Tee(buf1, buf2)
        tee.write("data")
        # StringIO flush doesn't change content, but we can call without error
        tee.flush()  # Should not raise

    def test_writable_returns_true_when_open(self):
        buf = StringIO()
        tee = _Tee(buf)
        self.assertTrue(tee.writable())

    def test_writable_raises_when_closed(self):
        buf = StringIO()
        tee = _Tee(buf)
        tee.close()
        with self.assertRaises(ValueError):
            tee.writable()

    def test_close_closes_secondary_files(self):
        buf1 = StringIO()
        buf2 = StringIO()
        tee = _Tee(buf1, buf2)
        tee.close()
        # buf2 should be closed
        with self.assertRaises(ValueError):
            buf2.write("x")


class TestFtee(unittest.TestCase):

    def test_ftee_redirects_stdout_and_writes_to_file(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            file_path = Path(tmpdir) / "out.txt"
            # Capture original stdout
            original_stdout = sys.stdout
            try:
                with ftee(file_path) as tee_stdout:
                    print("Hello Tee", end="")
                # After context, sys.stdout restored
                self.assertIs(sys.stdout, original_stdout)
                # File should contain the printed text
                content = file_path.read_text()
                self.assertIn("Hello Tee", content)
            finally:
                sys.stdout = original_stdout

    def test_ftee_writes_to_multiple_files(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            file1 = Path(tmpdir) / "out1.txt"
            file2 = Path(tmpdir) / "out2.txt"
            with ftee(file1, file2):
                print("Multi", end="")
            self.assertEqual(file1.read_text(), "Multi")
            self.assertEqual(file2.read_text(), "Multi")


class TestFteeExceptionSafety(unittest.TestCase):

    def test_ftee_restores_stdout_on_exception(self):
        original_stdout = sys.stdout
        with tempfile.TemporaryDirectory() as tmpdir:
            file_path = Path(tmpdir) / "out.txt"
            try:
                with ftee(file_path):
                    print("Before crash", end="")
                    raise RuntimeError("Simulated error")
            except RuntimeError:
                pass



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
import gc

import utlx
from utlx import const, weakconst


class TestConstDescriptors(unittest.TestCase):

    def test_const_returns_value(self):
        class Example:
            value = const(42)

        obj = Example()
        self.assertEqual(obj.value, 42)
        self.assertEqual(Example.value, 42)

    def test_const_is_readonly(self):
        class Example:
            value = const(100)

        obj = Example()
        with self.assertRaises(TypeError):
            obj.value = 200
        with self.assertRaises(TypeError):
            del obj.value

    def test_const_docstring(self):
        class Example:
            value = const(1, doc="This is a constant")

        self.assertEqual(Example.__dict__['value'].__doc__, "This is a constant")

    def test_weakconst_returns_value(self):
        class Ref:
            pass

        ref = Ref()

        class Example:
            value = weakconst(ref)

        obj = Example()
        self.assertIs(obj.value, ref)
        self.assertIs(Example.value, ref)

    def test_weakconst_is_readonly(self):
        class Ref:
            pass

        ref = Ref()

        class Example:
            value = weakconst(ref)

        obj = Example()
        with self.assertRaises(TypeError):
            obj.value = "new"
        with self.assertRaises(TypeError):
            del obj.value

    def test_weakconst_returns_none_after_gc(self):
        class Ref:
            pass

        ref = Ref()

        class Example:
            value = weakconst(ref)

        del ref
        gc.collect()

        obj = Example()
        self.assertIsNone(obj.value)

    def test_weakconst_docstring(self):
        class Ref:
            pass

        ref = Ref()

        class Example:
            value = weakconst(ref, doc="Weak reference")

        self.assertEqual(Example.__dict__['value'].__doc__, "Weak reference")



# Copyright (c) 1994 Adam Karpierz
# SPDX-License-Identifier: Zlib

from typing import Any
import ctypes

from ._detect import is_cpython

__all__ = ('from_oid',)

if is_cpython:

    def from_oid(oid: int | None, *, __ct: Any = ctypes) -> object | None:
        return __ct.cast(oid, __ct.py_object).value if oid else None

else:

    def from_oid(oid: int | None, *, __ct: Any = ctypes) -> object | None:
        from platform import python_implementation
        raise NotImplementedError("from_oid() currently works only on CPython!\n"
                                  f"Current interpreter: {python_implementation()}")

del Any, ctypes, is_cpython



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
from unittest import mock
import platform
import sys

import utlx
from utlx import module_path
from utlx.imports import import_file


class TestWindowsArchDetection(unittest.TestCase):

    def run_arch_test(self, machine_value, is_32bits, expected):
        with mock.patch("platform.machine", return_value=machine_value), \
             mock.patch("sys.maxsize", 2**32 if is_32bits else 2**64):
            import utlx.platform
            _arch = import_file(module_path(utlx.platform)/"windows/_arch.py", reload=True)
            result = _arch.get_python_arch()
            self.assertEqual(result, expected)

    def test_x86_64_variants(self):
        self.run_arch_test("x86_64", False, "x86_64")
        self.run_arch_test("x86_64", True, "x86")
        self.run_arch_test("amd64", False, "x86_64")
        self.run_arch_test("amd64", True, "x86")

    def test_x86_aliases(self):
        for variant in ["x86", "i386", "i686", "ia32"]:
            self.run_arch_test(variant, True, "x86")

    def test_arm_variants(self):
        self.run_arch_test("arm64", False, "arm64")
        self.run_arch_test("arm64", True, "arm")
        self.run_arch_test("armv7l", True, "arm")
        self.run_arch_test("aarch64", False, "arm64")
        self.run_arch_test("aarch64", True, "arm")

    def test_unknown_arch(self):
        self.run_arch_test("unknown", False, None)
        self.run_arch_test("weirdarch", True, None)


class TestLinuxArchDetection(unittest.TestCase):

    def run_arch_test(self, machine_value, is_32bits, expected,
                      fpu_present=None, byteorder="little"):
        with mock.patch("platform.machine", return_value=machine_value), \
             mock.patch("sys.maxsize", 2**32 if is_32bits else 2**64), \
             mock.patch("sys.byteorder", byteorder):
            import utlx.platform
            # optional mock of has_fpu()
            _arch = import_file(module_path(utlx.platform)/"linux/_arch.py", reload=True)
            if fpu_present is not None:
                has_fpu_saved, _arch.has_fpu = _arch.has_fpu, lambda: fpu_present
            try:
                result = _arch.get_python_arch()
            finally:
                if fpu_present is not None:
                    _arch.has_fpu = has_fpu_saved
            self.assertEqual(result, expected)

    def test_x86_variants(self):
        self.run_arch_test("x86_64", False, "x86_64")
        self.run_arch_test("x86_64", True, "x86")
        self.run_arch_test("amd64", False, "x86_64")
        self.run_arch_test("amd64", True, "x86")
        for variant in ["i386", "i686", "x86", "ia32"]:
            self.run_arch_test(variant, True, "x86")

    def test_arm64_variants(self):
        self.run_arch_test("aarch64", False, "aarch64")
        self.run_arch_test("arm64", False, "aarch64")
        self.run_arch_test("aarch64_be", False, "aarch64be", byteorder="big")
        # 32-bit mode on ARMv8 -> armhf/armle
        self.run_arch_test("aarch64", True, "armhf", fpu_present=True)
        self.run_arch_test("aarch64", True, "armle", fpu_present=False)

    def test_arm32_variants(self):
        # soft-float group
        for variant in ["armv6l", "armv5tel", "armv4tl", "armv4l", "arml", "armel"]:
            self.run_arch_test(variant, True, "armle")
        # hard-float group
        self.run_arch_test("armv7l", True, "armhf", fpu_present=True)
        self.run_arch_test("armv7l", True, "armle", fpu_present=False)
        self.run_arch_test("armv8l", True, "armhf", fpu_present=True)
        self.run_arch_test("armv8l", True, "armle", fpu_present=False)

    def test_arm_big_endian(self):
        self.run_arch_test("armv7b", True, None)
        self.run_arch_test("aarch64_be", True, None)

    def test_ppc_variants(self):
        # ppc64 big-endian
        self.run_arch_test("ppc64", False, "ppc64", byteorder="big")
        self.run_arch_test("powerpc64", False, "ppc64", byteorder="big")
        # ppc64 little-endian
        self.run_arch_test("ppc64le", False, "ppc64le", byteorder="little")
        self.run_arch_test("powerpc64le", False, "ppc64le", byteorder="little")
        # ppc64 32bit-mode,big-endian
        self.run_arch_test("ppc64", True, "ppc", byteorder="big")
        self.run_arch_test("powerpc64", True, "ppc", byteorder="big")
        # ppc64 32bit-mode,little-endian
        self.run_arch_test("ppc64le", True, "ppcle", byteorder="little")
        self.run_arch_test("powerpc64le", True, "ppcle", byteorder="little")
        # ppc32 big-endian
        self.run_arch_test("ppc", True, "ppc", byteorder="big")
        self.run_arch_test("powerpc", True, "ppc", byteorder="big")
        # ppc32 little-endian
        self.run_arch_test("ppcle", True, "ppcle", byteorder="little")
        self.run_arch_test("powerpcle", True, "ppcle", byteorder="little")

    def test_mips_variants(self):
        # MIPS64 big-endian
        self.run_arch_test("mips64", False, "mips64", byteorder="big")
        self.run_arch_test("mips64eb", False, "mips64", byteorder="big")
        # MIPS64 little-endian
        self.run_arch_test("mips64el", False, "mips64le", byteorder="little")
        self.run_arch_test("mips64le", False, "mips64le", byteorder="little")
        # MIPS64 32bit-mode,big-endian
        self.run_arch_test("mips64", True, "mips", byteorder="big")
        self.run_arch_test("mips64eb", True, "mips", byteorder="big")
        # MIPS64 32bit-mode,little-endian
        self.run_arch_test("mips64el", True, "mipsle", byteorder="little")
        self.run_arch_test("mips64le", True, "mipsle", byteorder="little")
        # MIPS32 big-endian
        self.run_arch_test("mips", True, "mips", byteorder="big")
        self.run_arch_test("mipseb", True, "mips", byteorder="big")
        # MIPS32 little-endian
        self.run_arch_test("mipsel", True, "mipsle", byteorder="little")
        self.run_arch_test("mipsle", True, "mipsle", byteorder="little")

    def test_other_architectures(self):
        self.run_arch_test("riscv64", False, "riscv64")
        self.run_arch_test("s390x", False, "s390x")

    def test_unknown_arch(self):
        self.run_arch_test("unknown", False, None)
        self.run_arch_test("weirdarch", True, None)


class TestMacOSArchDetection(unittest.TestCase):

    def run_arch_test(self, machine_value, is_32bits, expected):
        with mock.patch("platform.machine", return_value=machine_value), \
             mock.patch("sys.maxsize", 2**32 if is_32bits else 2**64):
            import utlx.platform
            _arch = import_file(module_path(utlx.platform)/"macos/_arch.py", reload=True)
            result = _arch.get_python_arch()
            self.assertEqual(result, expected)

    def test_x86_variants(self):
        self.run_arch_test("x86_64", False, "x86_64")
        self.run_arch_test("x86_64", True, "x86")
        self.run_arch_test("amd64", False, "x86_64")
        self.run_arch_test("amd64", True, "x86")
        for variant in ["i386", "i686", "x86", "ia32"]:
            self.run_arch_test(variant, True, "x86")

    def test_arm64_variants(self):
        self.run_arch_test("arm64", False, "arm64")
        self.run_arch_test("arm64", True, "arm")

    def test_unknown_arch(self):
        self.run_arch_test("unknown", False, None)
        self.run_arch_test("weirdarch", True, None)



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
import types

import utlx
from utlx import public, private


class TestPublicPrivate(unittest.TestCase):

    def setUp(self):
        global __all__
        __all__ = []

    def test_public_decorator_adds_to___all__(self):
        @public
        def sample(): return 1  # pragma: no cover
        self.assertIn("sample", __all__)

    def test_public_decorator_adds_to___all__repeated(self):
        @public
        def sample(): return 1  # pragma: no cover
        @public
        def sample(): return 1  # pragma: no cover
        self.assertIn("sample", __all__)

    def test_public_function_form_adds_single_name(self):
        result = public(test_func=lambda: 42)
        self.assertEqual(result(), 42)
        self.assertIn("test_func", __all__)
        self.assertTrue(callable(test_func))

    def test_public_function_form_adds_single_name_repeated(self):
        def test_func(): return 64
        result = public(test_func=test_func)
        result = public(test_func=test_func)
        self.assertEqual(result(), 64)
        self.assertIn("test_func", __all__)
        self.assertTrue(callable(test_func))

    def test_public_function_form_adds_multiple_names(self):
        result = public(a=lambda: 1, b=lambda: 2)
        self.assertIsInstance(result, tuple)
        self.assertEqual(result[0](), 1)
        self.assertEqual(result[1](), 2)
        self.assertIn("a", __all__)
        self.assertIn("b", __all__)

    def test_public_decorator_raises_on_kwargs(self):
        with self.assertRaises(AssertionError):
            public(lambda: 1, x=2)

    def test_private_removes_from___all__(self):
        @public
        def hidden(): return "secret"  # pragma: no cover
        self.assertIn("hidden", __all__)
        private(hidden)
        self.assertNotIn("hidden", __all__)

    def test___all___must_be_list(self):
        global __all__
        __all__ = "not a list"
        with self.assertRaises(TypeError):
            public(test=lambda: 1)
        def ghost(): return "boo"  # pragma: no cover
        with self.assertRaises(TypeError):
            private(ghost)

    def test_private_does_nothing_if_name_not_in___all__(self):
        def ghost(): return "boo"  # pragma: no cover
        private(ghost)  # Should not raise
        self.assertNotIn("ghost", __all__)



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
import sys
import platform
import types
import tempfile
import shutil
from pathlib import Path

import utlx
from utlx.imports import import_static, import_file
from utlx.platform import is_pypy


class TestImportStatic(unittest.TestCase):

    def test_import_builtin_module(self):
        mod = import_static("math")
        self.assertIsInstance(mod, types.ModuleType)
        self.assertEqual(mod.__name__, "math")

    def test_import_with_reload(self):
        mod1 = import_static("enum")
        mod2 = import_static("enum", reload=True)
        self.assertIsInstance(mod2, types.ModuleType)
        self.assertEqual(mod2.__name__, "enum")
        self.assertIsNot(mod1, mod2)

    def test_import_nonexistent_module(self):
        with self.assertRaises(ImportError):
            import_static("nonexistent_module_abcxyz")


class TestImportFile(unittest.TestCase):

    def setUp(self):
        self.temp_dir = Path(tempfile.mkdtemp())
        self.module_path = self.temp_dir/"testmod.py"
        self.module_path.write_text("x = 42\ny = 'hello'\n")

        # Package: directory with __init__.py
        self.pkg_dir = self.temp_dir/"mypkg"
        self.pkg_dir.mkdir()
        (self.pkg_dir/"__init__.py").write_text("a = 'package'\nb = 123\n")

        # Ensure temp_dir is in sys.path for strict_sys_path=True
        sys.path.insert(0, str(self.temp_dir))

    def tearDown(self):
        sys.path.remove(str(self.temp_dir))
        shutil.rmtree(self.temp_dir)

    def test_import_file_basic(self):
        mod = import_file(self.module_path)
        self.assertIsInstance(mod, types.ModuleType)
        self.assertEqual(mod.x, 42)
        self.assertEqual(mod.y, "hello")

    def test_import_file_with_custom_name(self):
        mod = import_file(self.module_path, name="custom_name")
        self.assertEqual(mod.__name__, "custom_name")

    def test_import_file_improper(self):
        module_path = self.temp_dir/"testmod.txt"
        module_path.write_text("\n")
        with self.assertRaises(ImportError):
            import_file(module_path)

    def test_import_file_reload(self):
        mod1 = import_file(self.module_path)
        mod2 = import_file(self.module_path, reload=True)
        self.assertIsNot(mod1, mod2)

    def test_import_file_strict_sys_path_violation(self):
        outside_path = Path(tempfile.gettempdir())/"outside.py"
        outside_path.write_text("z = 99\n")
        try:
            with self.assertRaises(ImportError):
                import_file(outside_path, strict_sys_path=True)
            # Default: strict_sys_path == True
            with self.assertRaises(ImportError):
                import_file(outside_path)
            import_file(outside_path, strict_sys_path=False)
        finally:
            outside_path.unlink()

    def test_import_file_nonexistent(self):
        with self.assertRaises(ImportError):
            import_file(self.temp_dir/"missing.py")

    def test_import_package_directory(self):
        mod = import_file(self.pkg_dir)
        self.assertIsInstance(mod, types.ModuleType)
        self.assertEqual(mod.a, "package")
        self.assertEqual(mod.b, 123)
        self.assertEqual(mod.__name__, "mypkg")

    def test_import_package_with_custom_name(self):
        mod = import_file(self.pkg_dir, name="custompkg")
        self.assertEqual(mod.__name__, "custompkg")

    def test_import_package_reload(self):
        mod1 = import_file(self.pkg_dir)
        mod2 = import_file(self.pkg_dir, reload=True)
        self.assertIsNot(mod1, mod2)



# flake8-in-file-ignores: noqa: F821

# Copyright (c) 2011 Adam Karpierz
# SPDX-License-Identifier: Zlib

# Strictly based on Thomas Heller's recipe.
# https://sourceforge.net/p/ctypes/mailman/message/22650014/

"""This module implements pointer arithmetic for ctypes pointers."""

from __future__ import annotations

from typing import SupportsIndex, Any
from ctypes import _Pointer as POINTER  # noqa: N814
import ctypes

__all__ = ('c_ptr_add', 'c_ptr_sub', 'c_ptr_iadd', 'c_ptr_isub', 'POINTER')


def c_ptr_add(ptr: POINTER[Any], other: SupportsIndex) -> POINTER[Any]:
    """
    Add an integer to a pointer instance.

    Returns a new pointer:

    >>> import ctypes
    >>> string_ptr = ctypes.c_char_p(b'foobar')
    >>> string_ptr.value
    b'foobar'
    >>> p2 = c_ptr_add(string_ptr, 3)
    >>> print(p2.value)
    b'bar'
    >>> string_ptr.value
    b'foobar'
    >>>
    """
    try:
        offset = other.__index__()
    except AttributeError:
        raise TypeError("Can only add integer to pointer")
    void_p = ctypes.cast(ptr, ctypes.c_void_p)
    void_p.value = (void_p.value or 0) + offset  # * ctypes.sizeof(ptr._type_)
    return ctypes.cast(void_p, type(ptr))


def c_ptr_sub(ptr: POINTER[Any], other: POINTER[Any] | SupportsIndex) -> POINTER[Any] | int:
    """
    Substract an integer or a pointer from a pointer.

    Returns a new pointer or an integer.

    >>> import ctypes
    >>> string_ptr = ctypes.c_char_p(b'foobar')
    >>> string_ptr.value
    b'foobar'
    >>> p2 = c_ptr_add(string_ptr, 3)
    >>> print(p2.value)
    b'bar'
    >>> string_ptr.value
    b'foobar'
    >>> print(c_ptr_sub(p2, string_ptr))
    3
    >>> print(c_ptr_sub(string_ptr, p2))
    -3
    >>>
    >>> p3 = c_ptr_sub(p2, 3)
    >>> p3.value
    b'foobar'
    >>>
    >>> c_ptr_sub(string_ptr, p3)
    0
    >>>
    """
    if isinstance(other, (POINTER, ctypes.c_void_p,
                          ctypes.c_char_p, ctypes.c_wchar_p)):
        if type(ptr) is not type(other):
            raise TypeError("Both pointers must be of the same type")
        return ((ctypes.cast(ptr,   ctypes.c_void_p).value or 0) -  # noqa: W504
                (ctypes.cast(other, ctypes.c_void_p).value or 0))
    else:
        try:
            offset = other.__index__()
        except AttributeError:
            raise TypeError("Can only substract pointer or integer from pointer")
        void_p = ctypes.cast(ptr, ctypes.c_void_p)
        void_p.value = (void_p.value or 0) - offset  # * ctypes.sizeof(ptr._type_)
        return ctypes.cast(void_p, type(ptr))


def c_ptr_iadd(ptr: POINTER[Any], other: SupportsIndex) -> None:
    """
    Add an integer to a pointer instance in place:

    >>> import ctypes
    >>> string_ptr = ctypes.c_char_p(b'foobar')
    >>> string_ptr.value
    b'foobar'
    >>> c_ptr_iadd(string_ptr, 3)
    >>> string_ptr.value
    b'bar'
    >>>
    """
    try:
        offset = other.__index__()
    except AttributeError:
        raise TypeError("Can only add integer to pointer")
    void_p = ctypes.cast(ctypes.pointer(ptr),
                         ctypes.POINTER(ctypes.c_void_p)).contents
    void_p.value = (void_p.value or 0) + offset  # * ctypes.sizeof(ptr._type_)


def c_ptr_isub(ptr: POINTER[Any], other: SupportsIndex) -> None:
    """
    Substract an integer or a pointer from a pointer.

    Returns a new pointer or an integer.

    >>> import ctypes
    >>> string_ptr = ctypes.c_char_p(b'foobar')
    >>> string_ptr.value
    b'foobar'
    >>> c_ptr_iadd(string_ptr, 4)
    >>> string_ptr.value
    b'ar'
    >>> c_ptr_isub(string_ptr, 2)
    >>> string_ptr.value
    b'obar'
    >>> c_ptr_isub(string_ptr, 1)
    >>> string_ptr.value
    b'oobar'
    >>> c_ptr_isub(string_ptr, 1)
    >>> string_ptr.value
    b'foobar'
    >>>
    """
    try:
        offset = other.__index__()
    except AttributeError:
        raise TypeError("Can only substract integer from pointer")
    void_p = ctypes.cast(ctypes.pointer(ptr),
                         ctypes.POINTER(ctypes.c_void_p)).contents
    void_p.value = (void_p.value or 0) - offset  # * ctypes.sizeof(ptr._type_)


del SupportsIndex, Any


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# Copyright (c) 1994 Adam Karpierz
# SPDX-License-Identifier: Zlib

import ctypes

__all__ = (
    'USHRT_MAX',  'SHRT_MAX',  'SHRT_MIN',
    'UINT_MAX',   'INT_MAX',   'INT_MIN',
    'ULONG_MAX',  'LONG_MAX',  'LONG_MIN',
    'ULLONG_MAX', 'LLONG_MAX', 'LLONG_MIN',
)

# include <limits.h>
USHRT_MAX  = ctypes.c_ushort(-1).value
SHRT_MAX   = USHRT_MAX >> 1
SHRT_MIN   = -SHRT_MAX - 1
UINT_MAX   = ctypes.c_uint(-1).value
INT_MAX    = UINT_MAX >> 1
INT_MIN    = -INT_MAX - 1
ULONG_MAX  = ctypes.c_ulong(-1).value
LONG_MAX   = ULONG_MAX >> 1
LONG_MIN   = -LONG_MAX - 1
ULLONG_MAX = ctypes.c_ulonglong(-1).value
LLONG_MAX  = ULLONG_MAX >> 1
LLONG_MIN  = -LLONG_MAX - 1

del ctypes



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
from unittest import mock
import sys
import os
import platform
import importlib

import utlx


class TestPlatformDetect(unittest.TestCase):

    def reload_platform(self, platform_attrs=None, sys_attrs=None, os_attrs=None):
        """Helper function for reloading the module with mocked values."""
        patches = []
        if platform_attrs:
            for name, value in platform_attrs.items():
                patches.append(mock.patch.object(platform, name, value, create=True))
        else: pass  # pragma: no cover
        if sys_attrs:
            for name, value in sys_attrs.items():
                patches.append(mock.patch.object(sys, name, value, create=True))
        else: pass  # pragma: no cover
        if os_attrs:
            for name, value in os_attrs.items():
                patches.append(mock.patch.object(os, name, value, create=True))
        else: pass  # pragma: no cover

        for p in patches:
            p.start()

        # Force a fresh import
        sys.modules.pop("utlx.platform._detect", None)
        import utlx.platform._detect as _detect
        _detect = importlib.reload(_detect)

        for p in reversed(patches):
            p.stop()

        return _detect

    def test_is_windows_true_for_win32(self):
        utlx_platform = self.reload_platform(sys_attrs={"platform": "win32"},
                                             platform_attrs={"win32_ver": lambda: ("10",)})
        self.assertTrue(utlx_platform.is_windows)

    def test_is_linux_true(self):
        utlx_platform = self.reload_platform(sys_attrs={"platform": "linux"},
                                             platform_attrs={"win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_linux)

    def test_is_macos_true(self):
        utlx_platform = self.reload_platform(sys_attrs={"platform": "darwin"},
                                             platform_attrs={"win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_macos)

    def test_is_bsd_true(self):
        utlx_platform = self.reload_platform(sys_attrs={"platform": "freebsd"},
                                             platform_attrs={"win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_bsd)

    def test_is_sunos_true(self):
        utlx_platform = self.reload_platform(sys_attrs={"platform": "sunos"},
                                             platform_attrs={"win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_sunos)

    def test_is_aix_true(self):
        utlx_platform = self.reload_platform(sys_attrs={"platform": "aix"},
                                             platform_attrs={"win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_aix)

    def test_is_android_true(self):
        utlx_platform = self.reload_platform(sys_attrs={"getandroidapilevel": lambda: 33},
                                             platform_attrs={"win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_android)

    def test_is_posix_true(self):
        utlx_platform = self.reload_platform(os_attrs={"name": "posix"},
                                             platform_attrs={"win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_posix)

    def test_is_32bits_true(self):
        utlx_platform = self.reload_platform(sys_attrs={"maxsize": 2**32},
                                             platform_attrs={"win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_32bits)

    def test_is_ucs2_true(self):
        utlx_platform = self.reload_platform(sys_attrs={"maxunicode": 0xFFFF},
                                             platform_attrs={"win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_ucs2)

    def test_is_cpython_true(self):
        utlx_platform = self.reload_platform(platform_attrs={"python_implementation": lambda: "CPython",
                                                             "win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_cpython)

    def test_is_pypy_true(self):
        utlx_platform = self.reload_platform(platform_attrs={"python_implementation": lambda: "PyPy",
                                                             "win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_pypy)

    def test_is_ironpython_true(self):
        utlx_platform = self.reload_platform(platform_attrs={"python_implementation": lambda: "IronPython",
                                                             "system": lambda: "CLI",
                                                             "win32_ver": lambda: ("",)},
                                             sys_attrs={"platform": "cli"})
        self.assertTrue(utlx_platform.is_ironpython)

    def test_is_wsl_true(self):
        Uname = type("Uname", (), {"system": platform.system(),
                                   "release": "5.15.0-microsoft-standard"})
        utlx_platform = self.reload_platform(platform_attrs={"uname": lambda: Uname(),
                                                             "win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_wsl)

    def test_is_cygwin_true(self):
        utlx_platform = self.reload_platform(sys_attrs={"platform": "cygwin"},
                                             platform_attrs={"win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_cygwin)

    def test_is_msys_true(self):
        utlx_platform = self.reload_platform(sys_attrs={"platform": "msys"},
                                             platform_attrs={"win32_ver": lambda: ("",)})
        self.assertTrue(utlx_platform.is_msys)



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
from unittest import mock
import types
import inspect
from pathlib import Path

import utlx
from utlx import module_path


class TestModulePath(unittest.TestCase):

    def test_module_path_with_explicit_module(self):
        # We use a built-in module as an example
        import os
        path = module_path(os)
        self.assertIsInstance(path, Path)
        self.assertTrue(path.exists())
        self.assertTrue(path.is_dir())

    def test_module_path_with_none_detects_current_file(self):
        # The function should return the path to the directory of this test file
        current_file = inspect.getfile(inspect.currentframe())
        expected_dir = Path(current_file).resolve().parent
        result = module_path()
        self.assertEqual(result, expected_dir)

    def test_module_path_with_level_argument(self):
        # We test whether the level works correctly in a deeper call stack
        def inner():
            return module_path(level=2)

        def outer():
            return inner()

        expected_dir = Path(inspect.getfile(inspect.currentframe())).resolve().parent
        self.assertEqual(outer(), expected_dir)

    @unittest.skip("test_module_path_fallback_to___file__: failed!")
    def test_module_path_fallback_to___file__(self):
        # We simulate the absence of a module in the frame, forcing the use of file
        frame = inspect.currentframe()
        fake_globals = {"__file__": __file__}
        with mock.patch("sys._getframe", return_value=frame), \
             mock.patch("inspect.getmodule", return_value=None), \
             mock.patch.object(frame, "f_globals", fake_globals):
            result = module_path(module=None)
            expected = Path(__file__).resolve().parent
            self.assertEqual(result, expected)


class TestModulePathMocked(unittest.TestCase):

    def test_module_explicit(self):
        dummy_module = mock.Mock(spec=types.ModuleType)
        with mock.patch("inspect.getfile", return_value="/fake/path/module.py"):
            result = module_path(dummy_module)
        self.assertEqual(result, Path("/fake/path").resolve())

    def test_module_none_with_module_detected(self):
        dummy_module = mock.Mock(spec=types.ModuleType)
        frame_mock = mock.Mock()
        frame_mock.f_globals = {"__file__": "/fake/path/from_frame.py"}

        with mock.patch("sys._getframe", return_value=frame_mock), \
             mock.patch("inspect.getmodule", return_value=dummy_module), \
             mock.patch("inspect.getfile", return_value="/fake/path/from_module.py"):
            result = module_path(None)
        self.assertEqual(result, Path("/fake/path").resolve())

    def test_module_none_without_module_detected(self):
        frame_mock = mock.Mock()
        frame_mock.f_globals = {"__file__": "/fake/path/from_globals.py"}

        with mock.patch("sys._getframe", return_value=frame_mock), \
             mock.patch("inspect.getmodule", return_value=None):
            result = module_path(None)
        self.assertEqual(result, Path("/fake/path").resolve())

    def test_module_none_with_level(self):
        frame_mock = mock.Mock()
        frame_mock.f_globals = {"__file__": "/fake/path/level_up.py"}

        with mock.patch("sys._getframe", return_value=frame_mock), \
             mock.patch("inspect.getmodule", return_value=None):
            result = module_path(level=1)
        self.assertEqual(result, Path("/fake/path").resolve())



# Read & write properties
#
# Copyright (c) 2006 by Philipp "philiKON" von Weitershausen
#                       philikon@philikon.de
#
# Freely distributable under the terms of the Zope Public License, v2.1.
#
# See _rwproperty.txt for detailed explanations

from typing import TypeAlias, Any
from collections.abc import Callable
import sys

__all__ = ('getproperty', 'setproperty', 'delproperty')

AnyCallable: TypeAlias = Callable[..., Any]


class _property:

    def __new__(cls, func: AnyCallable) -> property:  # type: ignore[misc]
        """Constructor"""
        # ugly, but common hack
        frame = sys._getframe(1)
        oldprop = frame.f_locals.get(func.__name__)
        if oldprop is None:
            return cls.create_property(func)
        elif isinstance(oldprop, property):
            return cls.enhance_property(oldprop, func)
        else:
            raise TypeError("read & write properties cannot be mixed with "
                            "other attributes except regular property objects.")

    @staticmethod
    def create_property(func: AnyCallable) -> property:
        raise NotImplementedError()

    @staticmethod
    def enhance_property(oldprop: property, func: AnyCallable) -> property:
        raise NotImplementedError()


class getproperty(_property):

    @staticmethod
    def create_property(func: AnyCallable) -> property:
        return property(func)

    @staticmethod
    def enhance_property(oldprop: property, func: AnyCallable) -> property:
        return property(func, oldprop.fset, oldprop.fdel)


class setproperty(_property):

    @staticmethod
    def create_property(func: AnyCallable) -> property:
        return property(None, func)

    @staticmethod
    def enhance_property(oldprop: property, func: AnyCallable) -> property:
        return property(oldprop.fget, func, oldprop.fdel)


class delproperty(_property):

    @staticmethod
    def create_property(func: AnyCallable) -> property:
        return property(None, None, func)

    @staticmethod
    def enhance_property(oldprop: property, func: AnyCallable) -> property:
        return property(oldprop.fget, oldprop.fset, func)


if __name__ == "__main__":
    import doctest
    doctest.testfile("_rwproperty.txt")



# Copyright (c) 2012 Adam Karpierz
# SPDX-License-Identifier: Zlib

from typing import TypeAlias
from os import PathLike
import sys
import types
import importlib.util
from pathlib import Path

__all__ = ('import_static', 'import_file')

StrPath: TypeAlias = str | PathLike[str]


def import_static(name: str, *, package: str | None = None,
                  reload: bool = False) -> types.ModuleType:
    """
    Import a module from the standard library or installed packages, \
    bypassing local shadowing and optionally forcing reload.

    Args:
        name (str): Fully qualified module name.
        package (str | None): Optional relative import context.
        reload (bool): If True, forces reloading the module.

    Returns:
        types.ModuleType: Imported module object.

    Raises:
        ImportError: If the module cannot be found or loaded.
    """
    spec = importlib.util.find_spec(name, package=package)
    if spec is None or spec.loader is None:
        raise ImportError(f"Module '{name}' not found")

    if not reload and spec.name in sys.modules:
        return sys.modules[spec.name]

    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


def import_file(path: StrPath, *,
                name: str | None = None, reload: bool = False,
                strict_sys_path: bool = True) -> types.ModuleType:
    """
    Import a module directly from a file path or package directory.

    Args:
        path (str | PathLike[str]): Path to .py file or package directory.
        name (str | None): Optional name to assign to the module.
        reload (bool): If True, forces reloading the module.
        strict_sys_path (bool): If True, ensures the file is within sys.path.

    Returns:
        types.ModuleType: Imported module object.

    Raises:
        ImportError: If the file does not exist, cannot be loaded,
                     or is outside sys.path when strict_sys_path is True.
    """
    path = Path(path).resolve()
    if not path.exists():
        raise ImportError(f"No module named '{path.stem}'")

    if strict_sys_path:
        if not any(path.is_relative_to(Path(p).resolve())
                   for p in sys.path if p):
            raise ImportError(f"Module path '{path}' is not within sys.path")

    if name is None:  name = path.stem
    if path.is_dir(): path = path/"__init__.py"

    spec = importlib.util.spec_from_file_location(name, str(path))
    if spec is None or spec.loader is None:
        raise ImportError(f"Cannot import module named '{name}'")

    if not reload and spec.name in sys.modules:
        return sys.modules[spec.name]

    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


del StrPath, PathLike



# Copyright (c) 2012 Adam Karpierz
# SPDX-License-Identifier: Zlib

import typing
from typing import Any
from collections.abc import Iterable, Sequence
try:
    from System import String  # type: ignore[import-not-found]
except ImportError:
    String = str

__all__ = ('issubtype', 'isiterable', 'issequence', 'remove_all',
           'print_refinfo')


def issubtype(x: Any, t: Any) -> bool:
    return isinstance(x, type) and issubclass(x, t)


def isiterable(x: Any) -> bool:
    return (isinstance(x, (Iterable, typing.Iterable))
            and not isinstance(x, (bytes, str, String)))


def issequence(x: Any) -> bool:
    return (isinstance(x, (Sequence, typing.Sequence))
            and not isinstance(x, (bytes, str, String)))


def remove_all(seq: list[Any], value: Any) -> None:
    seq[:] = (item for item in seq if item != value)


def print_refinfo(obj: Any) -> None:
    import sys
    type_name = getattr(type(obj), "__name__", getattr(
                getattr(obj, "__class__", None), "__name__",
                "???"))
    ref_count = ((sys.getrefcount(obj) - 2)
                 if hasattr(sys, "getrefcount") else None)
    print("Object info report",            file=sys.stderr)
    print("    obj type: ", type_name,     file=sys.stderr)
    print("    obj id:   ", id(obj),       file=sys.stderr)
    if ref_count is not None:
        print("    ref count:", ref_count, file=sys.stderr)



# flake8-in-file-ignores: noqa: A005,F401,F403,F405

# Copyright (c) 1994 Adam Karpierz
# SPDX-License-Identifier: Zlib

from ._detect import is_windows, is_linux, is_macos

if is_windows:  # pragma: no cover
    from .windows.capi import *  # type: ignore[assignment, unused-ignore]
elif is_linux:  # pragma: no cover
    from .linux.capi import *    # type: ignore[assignment, unused-ignore]
elif is_macos:  # pragma: no cover
    from .macos.capi import *    # type: ignore[assignment, unused-ignore]

del is_windows, is_linux, is_macos



# Copyright (c) 2012 Adam Karpierz
# SPDX-License-Identifier: Zlib

import typing
from typing import TypeVar, Generic, Any
from typing_extensions import Self
from collections.abc import Callable

__all__ = ('classproperty',)

_T = TypeVar("_T")
_C = TypeVar("_C", bound=type)


class classproperty(Generic[_C, _T]):

    __slots__ = ('_fget', 'fset', 'fdel', '__doc__')

    def __init__(self, fget: Callable[[_C], _T] | None = None,
                 fset: Any = None, fdel: Any = None,
                 doc: str | None = None) -> None:
        """Initializer"""
        self._fget = fget
        if fset is not None or fdel is not None:
            raise ValueError("classproperty only implements fget.")
        self.fset = fset
        self.fdel = fdel
        if doc is None and self._fget is not None:
            doc = self._fget.__doc__
        self.__doc__ = doc

    def __get__(self, this: Any, cls: _C | None = None) -> _T:
        """Access handler"""
        fget = self._fget
        if fget is None:
            raise AttributeError("unreadable attribute")
        if not callable(fget):
            raise TypeError("'{}' object is not callable".format(
                            type(fget).__name__))
        return fget(cls or typing.cast(_C, type(this)))

    @property
    def fget(self) -> Callable[[_C], _T] | None:
        return self._fget

    def getter(self, fget: Callable[[_C], _T] | None = None) -> Self:
        self._fget = fget
        return self


del TypeVar, Generic, Self, Callable



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
from unittest import mock
import sys
import io

import utlx
from utlx import issubtype, isiterable, issequence, remove_all, print_refinfo


class TestUtilityFunctions(unittest.TestCase):

    # --- issubtype ---
    def test_issubtype_with_valid_subclass(self):
        self.assertTrue(issubtype(int, object))

    def test_issubtype_with_non_type(self):
        self.assertFalse(issubtype(123, object))

    def test_issubtype_with_unrelated_types(self):
        self.assertFalse(issubtype(str, int))

    # --- isiterable ---
    def test_isiterable_with_list(self):
        self.assertTrue(isiterable([1, 2, 3]))

    def test_isiterable_with_tuple(self):
        self.assertTrue(isiterable((1, 2)))

    def test_isiterable_with_generator(self):
        self.assertTrue(isiterable((x for x in range(3))))

    def test_isiterable_with_string(self):
        self.assertFalse(isiterable("abc"))

    def test_isiterable_with_bytes(self):
        self.assertFalse(isiterable(b"abc"))

    def test_isiterable_with_custom_iterable(self):
        class Custom:
            def __iter__(self):
                return iter([1])
        seq = Custom()
        self.assertTrue(isiterable(seq))
        for item in seq:               # for full coverage
            self.assertEqual(item, 1)  #       -||-

    # --- issequence ---
    def test_issequence_with_list(self):
        self.assertTrue(issequence([1, 2]))

    def test_issequence_with_tuple(self):
        self.assertTrue(issequence((1, 2)))

    def test_issequence_with_string(self):
        self.assertFalse(issequence("text"))

    def test_issequence_with_bytes(self):
        self.assertFalse(issequence(b"data"))

    def test_issequence_with_custom_sequence(self):
        from collections.abc import Sequence
        class Custom(Sequence):
            def __getitem__(self, index):
                return index
            def __len__(self):
                return 12
        seq = Custom()
        self.assertTrue(issequence(seq))
        self.assertEqual(seq[7], 7)     # for full coverage
        self.assertEqual(len(seq), 12)  #       -||-

    # --- remove_all ---
    def test_remove_all_removes_all_occurrences(self):
        data = [1, 2, 3, 2, 4]
        remove_all(data, 2)
        self.assertEqual(data, [1, 3, 4])

    def test_remove_all_with_no_match(self):
        data = [1, 2, 3]
        remove_all(data, 99)
        self.assertEqual(data, [1, 2, 3])

    def test_remove_all_with_empty_list(self):
        data = []
        remove_all(data, 1)
        self.assertEqual(data, [])

    def test_remove_all_removes_everything(self):
        data = [5, 5, 5]
        remove_all(data, 5)
        self.assertEqual(data, [])


class TestPrintRefInfo(unittest.TestCase):

    def test_print_refinfo_outputs_expected_lines(self):
        obj = [1, 2, 3]
        captured = io.StringIO()
        sys.stderr = captured
        print_refinfo(obj)
        sys.stderr = sys.__stderr__
        output = captured.getvalue()
        self.assertIn("Object info report", output)
        self.assertIn("obj type:", output)
        self.assertIn("obj id:", output)

    def test_normal_object_with_refcount(self):
        obj = "hello"
        with mock.patch.object(sys, 'getrefcount', return_value=10, create=True), \
             mock.patch('sys.stderr', new_callable=io.StringIO) as fake_stderr:
            print_refinfo(obj)
            output = fake_stderr.getvalue()
            self.assertIn("Object info report", output)
            self.assertIn("obj type:", output)
            self.assertIn("obj id:", output)
            self.assertIn("ref count: 8", output)  # 10 - 2

    def test_object_without_getrefcount(self):
        obj = 123
        with mock.patch.object(sys, 'getrefcount', side_effect=AttributeError, create=True), \
             mock.patch('sys.stderr', new_callable=io.StringIO) as fake_stderr, \
             mock.patch('sys.hasattr', return_value=False):
            # We simulate the absence of getrefcount by manipulating hasattr
            original_hasattr = hasattr
            def fake_hasattr(obj, name):
                if obj is sys and name == "getrefcount":
                    return False
                return original_hasattr(obj, name)

            with mock.patch('builtins.hasattr', side_effect=fake_hasattr):
                hasattr(obj, "__doc__")  # for full coverage
                print_refinfo(obj)
                output = fake_stderr.getvalue()
                self.assertIn("Object info report", output)
                self.assertIn("obj type:", output)
                self.assertIn("obj id:", output)
                self.assertNotIn("ref count:", output)

    def test_object_with_missing_class_name(self):
        class Weird:
            def __getattribute__(self, name):
                if name in ['__class__']:
                    raise AttributeError("no __class__")
                return super().__getattribute__(name)

        obj = Weird()
        obj.__doc__  # for full coverage
        with mock.patch.object(sys, 'getrefcount', return_value=5, create=True), \
             mock.patch('sys.stderr', new_callable=io.StringIO) as fake_stderr:
            print_refinfo(obj)
            output = fake_stderr.getvalue()
            self.assertIn("Object info report", output)
            self.assertIn("obj type:", output)
            self.assertIn("ref count: 3", output)  # 5 - 2

    def test_object_with_missing_type_and_class_name(self):
        class Nameless:
            def __getattribute__(self, name):
                if name in ('__class__', '__name__'):
                    raise AttributeError("nope")
                return super().__getattribute__(name)

        obj = Nameless()
        obj.__doc__  # for full coverage
        with mock.patch.object(sys, 'getrefcount', return_value=7, create=True), \
             mock.patch('sys.stderr', new_callable=io.StringIO) as fake_stderr:
            print_refinfo(obj)
            output = fake_stderr.getvalue()
            self.assertIn("Object info report", output)
            self.assertIn("obj type:", output)
            self.assertIn("ref count: 5", output)



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest

import utlx
from utlx import defined


class TestDefinedFunction(unittest.TestCase):

    def test_defined_local_variable(self):
        def func():
            local_var = 123
            return defined("local_var")
        self.assertTrue(func())

    def test_defined_global_variable(self):
        global_var = "hello"
        self.assertTrue(defined("global_var"))

    def test_defined_missing_variable(self):
        self.assertFalse(defined("nonexistent_var"))

    def test_defined_shadowed_variable(self):
        # The variable exists only locally, not globally
        def func():
            shadow = "local"
            return defined("shadow")
        self.assertTrue(func())
        self.assertFalse("shadow" in globals())

    def test_defined_with_builtin(self):
        # Built-in functions are not treated as local or global
        self.assertFalse(defined("print"))  # print is in builtins, not in globals



# Copyright (c) 2016 Adam Karpierz
# SPDX-License-Identifier: Zlib

from typing import Any
from collections.abc import Sequence, Generator

__all__ = ('renumerate',)


def renumerate(sequence: Sequence[Any],
               start: int | None = None,
               end:   int | None = None) -> Generator[tuple[int, Any]]:
    """Reverse iterator for index, value of sequence.

    renumerate(sequence[, start]|[, end]) -> reverse iterator

    Return an enumerate object.  sequence must be another object that has a
    __reversed__() method or supports the sequence protocol (the __len__()
    method and the __getitem__() method with integer arguments starting at 0).
    The enumerate object yields pairs containing a count (from start, which
    defaults to len(sequence) - 1 or ends at end, which defaults to zero - but
    not both) and a value yielded by the sequence argument in reverse order.
    renumerate is useful for obtaining an indexed list in reverse order:
        (2, seq[2]), (1, seq[1]), (0, seq[0]), ...
    """
    if start is not None and end is not None:
        raise TypeError("renumerate() only accepts start argument or "
                        "end argument - not both.")
    if start is None:
        start = len(sequence) - 1
    if end is None:
        end = 0
    return ((-idx, elem) for idx, elem in enumerate(reversed(sequence),
                                                    start=-(start + end)))



# Copyright (c) 2018 Adam Karpierz
# SPDX-License-Identifier: Zlib

from ctypes.wintypes import HMODULE
from pathlib import Path

__all__ = ('dll_path', 'python_dll_path')


def dll_path(handle: HMODULE | int) -> Path | None:
    """Retrieves the fully qualified path for the file that contains the specified module.

    The module must have been loaded by the current process.
    """
    import ctypes
    from ctypes.wintypes import HMODULE, LPWSTR, DWORD
    MAX_PATH = 520
    GetModuleFileNameW = ctypes.windll.kernel32.GetModuleFileNameW
    GetModuleFileNameW.restype  = DWORD
    GetModuleFileNameW.argtypes = [HMODULE, LPWSTR, DWORD]
    buf = ctypes.create_unicode_buffer(MAX_PATH)
    # print("LENNNN", len(buf))
    result = GetModuleFileNameW(handle, buf, len(buf))
    dll_path = buf.value
    # print("@@@@@@@@@@@", handle, result, dll_path)
    return (Path(dll_path) if handle != 0 and result != 0
            and dll_path and Path(dll_path).exists() else None)


def python_dll_path() -> Path | None:
    """Retrieves the fully qualified path for the file that contains Python dll module.

    The module must have been loaded by the current process.
    """
    try:
        from ctypes import pythonapi
    except ImportError:  # pragma: no cover
        from sys import dllhandle
    else:
        dllhandle = pythonapi._handle
    return dll_path(dllhandle)


del HMODULE



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
import warnings

import utlx
from utlx import deprecated


class TestDeprecatedDecorator(unittest.TestCase):

    def test_deprecated_emits_warning(self):
        @deprecated
        def old_function():
            return "ok"

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = old_function()
            self.assertEqual(result, "ok")
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, DeprecationWarning))
            self.assertIn("Call to deprecated function 'old_function'", str(w[0].message))

    def test_deprecated_preserves_function_metadata(self):
        @deprecated
        def sample():
            """This is a sample function"""
            return 123

        self.assertEqual(sample.__name__, "sample")
        self.assertEqual(sample.__doc__, "This is a sample function")

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = sample()
            self.assertEqual(result, 123)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, DeprecationWarning))
            self.assertIn("Call to deprecated function 'sample'", str(w[0].message))

    def test_deprecated_with_arguments(self):
        @deprecated
        def add(a, b):
            return a + b

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = add(2, 3)
            self.assertEqual(result, 5)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, DeprecationWarning))
            self.assertIn("Call to deprecated function 'add'", str(w[0].message))



# flake8-in-file-ignores: noqa: A005,F401,F403,F405

# Copyright (c) 1994 Adam Karpierz
# SPDX-License-Identifier: Zlib

from ._detect import *
from ._oid    import *
from . import _limits as limits
if is_windows:  # pragma: no cover
    from .windows import arch as _arch
elif is_linux:  # pragma: no cover
    from .linux import arch as _arch
elif is_macos:  # pragma: no cover
    from .macos import arch as _arch
from . import capi
arch = _arch



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
import copy

import utlx
from utlx import adict, defaultadict


class TestAdict(unittest.TestCase):

    def test_item_assignment(self):
        d = adict()
        self.assertNotIn("x", d)
        d["x"] = 10
        self.assertEqual(d.x, 10)
        self.assertEqual(d["x"], 10)

    def test_attribute_assignment(self):
        d = adict()
        self.assertNotIn("x", d)
        d.x = 20
        self.assertEqual(d["x"], 20)
        self.assertEqual(d.x, 20)

    def test_missing_item_raises(self):
        d = adict()
        with self.assertRaises(KeyError):
            _ = d["missing"]
        with self.assertRaises(AttributeError):
            _ = d["missing"]

    def test_missing_attribute_raises(self):
        d = adict()
        with self.assertRaises(KeyError):
            _ = d.missing
        with self.assertRaises(AttributeError):
            _ = d.missing

    def test_item_deletion(self):
        d = adict(x=5)
        self.assertIn("x", d)
        del d["x"]
        self.assertNotIn("x", d)

    def test_attribute_deletion(self):
        d = adict(x=5)
        self.assertIn("x", d)
        del d.x
        self.assertNotIn("x", d)

    def test_missing_item_deletion_raises(self):
        d = adict()
        with self.assertRaises(KeyError):
            del d["missing"]
        with self.assertRaises(AttributeError):
            del d["missing"]

    def test_missing_attribute_deletion_raises(self):
        d = adict()
        with self.assertRaises(KeyError):
            del d.missing
        with self.assertRaises(AttributeError):
            del d.missing

    def test_self_copy_returns_new_instance(self):
        d1 = adict(a=1, b=2)
        d2 = d1.copy()
        self.assertIsInstance(d1, adict)
        self.assertIsInstance(d2, adict)
        self.assertEqual(d2, d1)
        self.assertIsNot(d1, d2)

    def test_copy_returns_new_instance(self):
        d1 = adict(a=1, b=2)
        d2 = copy.copy(d1)
        self.assertIsInstance(d1, adict)
        self.assertIsInstance(d2, adict)
        self.assertEqual(d2, d1)
        self.assertIsNot(d1, d2)

    def test_fromkeys_creates_instance(self):
        d = adict.fromkeys(["a", "b"], 0)
        self.assertIsInstance(d, adict)
        self.assertEqual(d["a"], 0)
        self.assertEqual(d.a, 0)
        self.assertEqual(d["b"], 0)
        self.assertEqual(d.b, 0)


class TestDefaultAdict(unittest.TestCase):

    def test_default_factory_behavior(self):
        d = defaultadict(lambda: "default")
        self.assertEqual(d["missing"], "default")

    def test_item_access_and_assignment(self):
        d = defaultadict(lambda: 0)
        self.assertNotIn("x", d)
        d["x"] = 100
        self.assertEqual(d["x"], 100)
        self.assertEqual(d.x, 100)

    def test_attribute_access_and_assignment(self):
        d = defaultadict(lambda: 0)
        self.assertNotIn("x", d)
        d.x = 100
        self.assertEqual(d["x"], 100)
        self.assertEqual(d.x, 100)

    def test_item_deletion(self):
        d = defaultadict(lambda: None, x=1)
        self.assertIn("x", d)
        del d["x"]
        self.assertNotIn("x", d)

    def test_attribute_deletion(self):
        d = defaultadict(lambda: None, x=1)
        self.assertIn("x", d)
        del d.x
        self.assertNotIn("x", d)

    def test_self_copy_returns_new_instance(self):
        d1 = defaultadict(None, a=1, b=2)
        d2 = d1.copy()
        self.assertIsInstance(d1, defaultadict)
        self.assertIsInstance(d2, defaultadict)
        self.assertEqual(d2, d1)
        self.assertIsNot(d1, d2)

    def test_copy_returns_new_instance(self):
        d1 = defaultadict(None, a=1, b=2)
        d2 = copy.copy(d1)
        self.assertIsInstance(d1, defaultadict)
        self.assertIsInstance(d2, defaultadict)
        self.assertEqual(d2, d1)
        self.assertIsNot(d1, d2)

    def test_self_copy_preserves_factory(self):
        d1 = defaultadict(lambda: 42, a=1)
        d2 = d1.copy()
        self.assertIsInstance(d1, defaultadict)
        self.assertIsInstance(d2, defaultadict)
        self.assertEqual(d2, d1)
        self.assertIsNot(d1, d2)
        self.assertEqual(d2["a"], 1)
        self.assertEqual(d2.a, 1)
        self.assertEqual(d2["missing"], 42)

    def test_copy_preserves_factory(self):
        d1 = defaultadict(lambda: 42, a=1)
        d2 = copy.copy(d1)
        self.assertIsInstance(d1, defaultadict)
        self.assertIsInstance(d2, defaultadict)
        self.assertEqual(d2, d1)
        self.assertIsNot(d1, d2)
        self.assertEqual(d2["a"], 1)
        self.assertEqual(d2.a, 1)
        self.assertEqual(d2["missing"], 42)

    def test_fromkeys_creates_instance(self):
        d = defaultadict.fromkeys(["a", "b"], "Y")
        # d.default_factory = lambda: "X"
        self.assertIsInstance(d, defaultadict)
        self.assertEqual(d["a"], "Y")
        self.assertEqual(d.a, "Y")
        self.assertEqual(d["b"], "Y")
        self.assertEqual(d.b, "Y")
        # self.assertEqual(d.missing, "X")
        # self.assertEqual(d["missing"], "X")

    def test_missing_attribute_raises(self):
        d = defaultadict(lambda: None)
        self.assertNotIn("nonexistent", d)
        item = d["nonexistent"]
        self.assertIsNone(item)
        item = d.nonexistent
        self.assertIsNone(item)
        item = d.__getitem__("nonexistent")
        self.assertIsNone(item)
        item = d.__getattr__("nonexistent")
        self.assertIsNone(item)



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest

import utlx
from utlx import unique, iter_unique


class TestUniqueFunctions(unittest.TestCase):

    def test_unique_with_integers(self):
        self.assertEqual(unique([1, 2, 2, 3, 1]), [1, 2, 3])

    def test_unique_with_strings(self):
        self.assertEqual(unique("banana"), ['b', 'a', 'n'])

    def test_unique_with_empty_iterable(self):
        self.assertEqual(unique([]), [])

    def test_unique_with_mixed_types(self):
        self.assertEqual(unique([1, "1", 1.0]), [1, "1"])

    def test_iter_unique_basic(self):
        result = list(iter_unique("AAAABBBCCDAABBB"))
        self.assertEqual(result, ['A', 'B', 'C', 'D'])

    def test_iter_unique_with_key(self):
        result = list(iter_unique("ABBCcAD", key=str.lower))
        self.assertEqual(result, ['A', 'B', 'C', 'D'])

    def test_iter_unique_empty(self):
        result = list(iter_unique([], key=str.lower))
        self.assertEqual(result, [])

    def test_iter_unique_custom_key(self):
        data = ["apple", "Apple", "banana", "Banana"]
        result = list(iter_unique(data, key=str.lower))
        self.assertEqual(result, ["apple", "banana"])

    def test_iter_unique_with_numbers_and_key(self):
        data = [1, 2, -1, -2, 3]
        result = list(iter_unique(data, key=abs))
        self.assertEqual(result, [1, 2, 3])

    def test_iter_unique_generator_behavior(self):
        gen = iter_unique([1, 2, 2, 3])
        self.assertTrue(hasattr(gen, '__iter__'))
        self.assertTrue(hasattr(gen, '__next__'))
        self.assertEqual(next(gen), 1)
        self.assertEqual(next(gen), 2)
        self.assertEqual(next(gen), 3)
        with self.assertRaises(StopIteration):
            next(gen)



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
import doctest
import ctypes as ct

from utlx import ctypes as ctx


class TestUtlxCtypes(unittest.TestCase):

    def test_utlx_ctypes_c_ptr(self):
        from utlx.ctypes import _c_ptr
        result = doctest.testmod(_c_ptr)
        if result.failed:  # pragma: no cover
            doctest.testmod(_c_ptr, raise_on_error=True)

    def test_invalid_c_ptr_parameters(self):
        obj = ct.c_char_p(b"ABCD")
        with self.assertRaisesRegex(TypeError, "Can only "):
            _ = ctx.c_ptr_add(obj,  "invalid because string")
        with self.assertRaisesRegex(TypeError, "Can only "):
            _ = ctx.c_ptr_sub(obj,  "invalid because string")
        with self.assertRaisesRegex(TypeError, "Both pointers must "):
            _ = ctx.c_ptr_sub(obj,  ct.c_void_p(345))
        with self.assertRaisesRegex(TypeError, "Can only "):
            ctx.c_ptr_iadd(obj, "invalid because string")
        with self.assertRaisesRegex(TypeError, "Can only "):
            ctx.c_ptr_isub(obj, "invalid because string")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
import types

import utlx
from utlx import renumerate


class TestRenumerate(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.seasons = ["Spring", "Summer", "Fall", "Winter"]

    def test_renumerate_type(self):
        self.assertIs(type(renumerate(self.seasons)), types.GeneratorType)
        self.assertIsInstance(renumerate(self.seasons), types.GeneratorType)
        for item in renumerate(self.seasons):
            self.assertIsInstance(item, tuple)
            self.assertEqual(len(item), 2)
            self.assertIs(type(item[0]), int)

    def test_renumerate_as_tuple(self):
        self.assertTupleEqual(tuple(renumerate(self.seasons)),
                              ((3, "Winter"), (2, "Fall"), (1, "Summer"), (0, "Spring")))
        self.assertTupleEqual(tuple(renumerate(self.seasons, start=4)),
                              ((4, "Winter"), (3, "Fall"), (2, "Summer"), (1, "Spring")))
        self.assertTupleEqual(tuple(renumerate(self.seasons, end=2)),
                              ((5, "Winter"), (4, "Fall"), (3, "Summer"), (2, "Spring")))
        with self.assertRaises(TypeError):
            renumerate(self.seasons, start=4, end=2)

    def test_renumerate_as_list(self):
        self.assertListEqual(list(renumerate(self.seasons)),
                             [(3, "Winter"), (2, "Fall"), (1, "Summer"), (0, "Spring")])
        self.assertListEqual(list(renumerate(self.seasons, start=4)),
                             [(4, "Winter"), (3, "Fall"), (2, "Summer"), (1, "Spring")])
        self.assertListEqual(list(renumerate(self.seasons, end=2)),
                             [(5, "Winter"), (4, "Fall"), (3, "Summer"), (2, "Spring")])
        with self.assertRaises(TypeError):
            renumerate(self.seasons, start=4, end=2)



# Copyright (c) 2004 Adam Karpierz
# SPDX-License-Identifier: Zlib

import unittest
from unittest import mock
import io
import subprocess
import logging

import utlx
from utlx import _run, run


class TestRunFunction(unittest.TestCase):

    def setUp(self):
        # Capture logs for verification
        self.log_stream = logging.getLogger(_run.__name__)
        self.log_stream.setLevel(logging.DEBUG)
        self.log_buffer = io.StringIO()
        self.log_handler = logging.StreamHandler(self.log_buffer)
        self.log_stream.addHandler(self.log_handler)
        self.addCleanup(self.log_stream.removeHandler, self.log_handler)

    def test_run_basic_command(self):
        """run() should call subprocess.run with given args and return CompletedProcess."""
        fake_cp = subprocess.CompletedProcess(args=["echo", "hi"], returncode=0)
        with mock.patch("subprocess.run", return_value=fake_cp) as mock_run:
            result = run("echo", "hi")
        mock_run.assert_called_once_with(["echo", "hi"], check=True)
        self.assertIs(result, fake_cp)

    def test_run_with_check_false(self):
        """run() should pass check=False when specified."""
        fake_cp = subprocess.CompletedProcess(args=["cmd"], returncode=1)
        with mock.patch("subprocess.run", return_value=fake_cp) as mock_run:
            result = run("cmd", check=False)
        mock_run.assert_called_once_with(["cmd"], check=False)
        self.assertIs(result, fake_cp)

    def test_run_masks_safestring_in_log(self):
        """SafeString arguments should be masked in debug log output."""
        fake_cp = subprocess.CompletedProcess(args=["secret"], returncode=0)
        safe_arg = run.SafeString("secret")
        with mock.patch("subprocess.run", return_value=fake_cp):
            run("echo", safe_arg)
        # Retrieve log output
        self.log_handler.flush()
        log_output = self.log_buffer.getvalue()
        self.assertIn("*****", log_output)

    def test_run_converts_all_args_to_str(self):
        """Non-string args should be converted to strings before passing to subprocess.run."""
        fake_cp = subprocess.CompletedProcess(args=["1", "2"], returncode=0)
        with mock.patch("subprocess.run", return_value=fake_cp) as mock_run:
            run(1, 2)
        called_args, called_kwargs = mock_run.call_args
        self.assertEqual(called_args[0], ["1", "2"])

    def test_constants_and_types(self):
        """Check that run.* constants and types are correctly assigned."""
        self.assertIs(run.PIPE, subprocess.PIPE)
        self.assertIs(run.STDOUT, subprocess.STDOUT)
        self.assertIs(run.DEVNULL, subprocess.DEVNULL)
        self.assertIs(run.CompletedProcess, subprocess.CompletedProcess)
        self.assertIs(run.SubprocessError, subprocess.SubprocessError)
        self.assertIs(run.TimeoutExpired, subprocess.TimeoutExpired)
        self.assertIs(run.CalledProcessError, subprocess.CalledProcessError)
        self.assertTrue(issubclass(run.SafeString, str))


class TestSplitKwargs(unittest.TestCase):

    def test_split_kwargs_basic(self):
        """split_kwargs should separate allowed and reserved keys."""
        kwargs = {"a": 1, "b": 2, "c": 3}
        forbidden = ["b", "c"]
        allowed, reserved = run.split_kwargs(kwargs, forbidden)
        self.assertEqual(allowed, {"a": 1})
        self.assertEqual(reserved, {"b": 2, "c": 3})

    def test_split_kwargs_no_forbidden(self):
        """If no forbidden keys, all kwargs should be allowed."""
        kwargs = {"x": 10}
        allowed, reserved = run.split_kwargs(kwargs, [])
        self.assertEqual(allowed, {"x": 10})
        self.assertEqual(reserved, {})

    def test_split_kwargs_all_forbidden(self):
        """If all keys are forbidden, allowed should be empty."""
        kwargs = {"x": 1, "y": 2}
        allowed, reserved = run.split_kwargs(kwargs, ["x", "y"])
        self.assertEqual(allowed, {})
        self.assertEqual(reserved, {"x": 1, "y": 2})
